# GloFlow application and media management/publishing platform
# Copyright (C) 2020 Ivan Trajkovic
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

import tensorflow as tf

from tensorflow.keras import layers, models
import matplotlib.pyplot as plt

import gf_ml_data

#----------------------------------------------
def main():

    # GENERATE_DATA
    generated__dataset = gf_ml_data.load__generated(p_generate_bool=True)

    # DATA_INPUT
    data_map = gf_ml_data.load__cifar10()

    # CREATE
    model = create(data_map)

    # FIT
    train_tf_history = fit(data_map, model)

    # EVALUATE
    evaluate(data_map,
        train_tf_history,
        model)

#----------------------------------------------
# CREATE

def create(p_data_map):

    print(" --- CREATE MODEL ")
    # model = models.Sequential()

    print(p_data_map["train_images"][:1].shape)

    #---------------------------
    # CONV_1
    # model.add(layers.Conv2D(32, (3, 3), activation="relu", input_shape=(32, 32, 3)))
    # model.add(layers.MaxPooling2D((2, 2)))

    input_l    = layers.Input(shape=(32, 32, 3))
    conv1_l    = layers.Conv2D(32, (3, 3), activation="relu")(input_l)
    maxpool1_l = layers.MaxPooling2D((2, 2))(conv1_l)

    print(maxpool1_l)
    
    #---------------------------
    # CONV_2
    # model.add(layers.Conv2D(64, (3, 3), activation="relu"))
    # model.add(layers.MaxPooling2D((2, 2)))

    conv2_l    = layers.Conv2D(64, (3, 3), activation="relu")(maxpool1_l)
    maxpool2_l = layers.MaxPooling2D((2, 2))(conv2_l)

    #---------------------------
    # # CONV_3
    # model.add(layers.Conv2D(128, (3, 3), activation="relu"))
    # model.add(layers.MaxPooling2D((2, 2)))
    # 
    # model.add(layers.Conv2D(128, (2, 2), activation="relu"))
    # model.add(layers.MaxPooling2D((2, 2)))
    # 
    # CONV_4
    # model.add(layers.Conv2D(256, (2, 2), activation="relu"))
    # model.add(layers.MaxPooling2D((2, 2)))

    #---------------------------
    # CONV_3
    # model.add(layers.Conv2D(64, (3, 3), activation="relu"))
    conv3_l = layers.Conv2D(64, (3, 3), activation="relu")(maxpool2_l)

    #---------------------------
    # FLATTEN
    # model.add(layers.Flatten())
    flatten_l = layers.Flatten()(conv3_l)

    #---------------------------
    # DENSE - 64
    # model.add(layers.Dense(64, activation="relu"))
    dense1_l = layers.Dense(64, activation="relu")(flatten_l)
    
    #---------------------------
    # DENSE - 10
    # model.add(layers.Dense(10))
    dense2_l = layers.Dense(10)(dense1_l)
    
    #---------------------------
    model = tf.keras.models.Model(inputs=input_l, outputs=dense2_l)
    model.summary()

    # COMPILE
    model.compile(optimizer = "adam",
        loss    = tf.keras.losses.SparseCategoricalCrossentropy(from_logits=True),
        metrics = ["accuracy"])

    return model

#----------------------------------------------
# FIT
def fit(p_data_map,
    p_model):

    print(" --- FIT MODEL ")
    epochs_num_int   = 1
    examples_num_int = 100


    print("run...")
    train_tf_history = p_model.fit(p_data_map["train_images"][:examples_num_int],
        p_data_map["train_labels"][:examples_num_int],
        epochs          = epochs_num_int,
        validation_data = (p_data_map["test_images"][:examples_num_int], p_data_map["test_labels"][:examples_num_int]))


    print(train_tf_history)


    for v in p_model.trainable_weights:
        print(v)

        print(v.shape)



    x = p_data_map["train_images"][0]



    import numpy as np

    # add one dimension to the tensor, to be (w, h, 3) -> (1, w, h, 3) 
    x_expanded = np.expand_dims(x, 0)
    y = p_model.predict(x_expanded)

    print(y)
    exit()

    return train_tf_history

#----------------------------------------------
# EVALUATE
def evaluate(p_data_map,
    p_train_tf_history,
    p_model):



    print(p_train_tf_history.history.keys())



    plt.plot(p_train_tf_history.history["acc"],     label = "accuracy")
    plt.plot(p_train_tf_history.history["val_acc"], label = "val_accuracy")
    plt.xlabel("Epoch")
    plt.ylabel("Accuracy")
    plt.ylim([0.5, 1])
    plt.legend(loc="lower right")

    test_loss, test_acc = p_model.evaluate(p_data_map["test_images"],
        p_data_map["test_labels"],
        verbose=2)





    print(test_loss)
    print(test_acc)



    plt.show()

#----------------------------------------------
main()


# GloFlow application and media management/publishing platform
# Copyright (C) 2024 Ivan Trajkovic
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

import json
import requests

import sentry_sdk

#--------------------------------------------
def run_llm(p_prompt_str,
	p_openai_api_key_str,
	p_model_str="gpt-4",
	p_temperature=0.7,
	p_functions_meta_map=None,
	p_verbose_bool=False):

	ctx_lst = []
	
	init_resp_str, first_func_call_map, sub_ctx_lst = model_request(p_prompt_str,
		p_openai_api_key_str,
		p_model_str=p_model_str,
		p_functions_meta_map = p_functions_meta_map)

	ctx_lst.extend(sub_ctx_lst)

	#--------------------------------------------
	def develop_context_with_model(p_first_func_call_map,
		p_ctx_map):

		if p_verbose_bool:
			print(">>>>>>>>>>>>>>>> entering function calls loop...")

		func_call_to_run_map = p_first_func_call_map

		i=0
		while True:

			if p_verbose_bool:
				print(f">>>>>>>>>>>>>>>> ---------------------------- local function execution {i}...")
			
			# dont run more than 5 function calls
			if i > 5:
				break
			
			#----------------------
			# FUNC_EXEC
			gpt_result_msg_map = func_exec(func_call_to_run_map,
				p_functions_meta_map)

			# add the function call to the conversation context
			p_ctx_map.append(gpt_result_msg_map)

			# print(gpt_result_msg_map)

			#----------------------
			# RETURN_RESULTS_TO_MODEL

			new_gen_text_str, new_func_call_map, sub_ctx_lst = model_request_with_ctx(p_ctx_map,
				p_openai_api_key_str,
				p_model_str=p_model_str,
				p_temperature=p_temperature,
				p_functions_meta_map=p_functions_meta_map,
				p_verbose_bool=p_verbose_bool)
			
			if new_func_call_map == None:
				
				# no more function calls needed by the model,
				# just return the final generated text
				return new_gen_text_str
			
			else:

				# there are more function calls to run, so run them
				func_call_to_run_map = new_func_call_map
				p_ctx_map.extend(sub_ctx_lst)

			#----------------------

			i+=1
	
	#--------------------------------------------

	if p_verbose_bool:

		print("first llm response --------------")
		print(init_resp_str)

	
	# FUNCTION_CALL
	# first reply is a function call that might result in a whole sequence of calls that a model
	# has to execute in order to gather all the context it needs to generate the final answer.
	if not first_func_call_map == None:
		
		develop_context_with_model(first_func_call_map, ctx_lst)

	# REGULAR_REPLY       
	else:

		print(init_resp_str)
		
		if not init_resp_str == "":
			result_map = json.loads(init_resp_str)

			if p_verbose_bool:
				print("LLM result:")
				print(result_map)

			return result_map
		else:
			return {}

#--------------------------------------------
def func_exec(p_func_call_map,
	p_functions_meta_map,
	p_json_output_bool=False):

	func_name_str = p_func_call_map["name"]

	# FUNC_ARGS - argument values are defined as a JSON map
	func_arguments_map = json.loads(p_func_call_map["arguments"])

	
	#----------------------
	# CALL_FUNCTION
	# apply map of arguments to the function as named arguments

	func_result = p_functions_meta_map[func_name_str]["func"](**func_arguments_map)

	#----------------------

	output_str = ""

	if p_json_output_bool:
		output_str = json.dumps(func_result)
	else:
		if not func_result == None:
			assert isinstance(func_result, str)
			output_str = func_result

	gpt_result_msg_map = {

		"role": "function",
		"name": func_name_str,
		"content": output_str
	}

	

	return gpt_result_msg_map

#--------------------------------------------
def model_request_with_ctx(p_ctx_map,
	p_api_key_str,
	p_model_str="gpt-4",
	p_temperature=0.7,
	p_functions_meta_map=None,
	p_verbose_bool=False):

	if p_verbose_bool:
		print(">>>>>>>>>>>>>>>> model request (with ctx)...")

	# GPT_MESSAGE_HISTORY
	messages_lst = p_ctx_map

	data_map = {
		"model": p_model_str,
		"messages": messages_lst,
		"temperature": p_temperature
	}

	# if function signatures metadata is supplied, pass it to the model
	# for it to be able to use it in the completion.
	if not p_functions_meta_map == None:

		# definitions of functions meant for the GPT model
		functions_gpt_defs_lst = [v["gpt_def_map"] for _, v in p_functions_meta_map.items()]

		data_map["functions"] = functions_gpt_defs_lst

	# HTTP_REQUEST
	r_map = make_request(data_map, p_api_key_str, p_verbose_bool=p_verbose_bool)

	#----------------------
	# MSG

	msg_map = r_map["choices"][0]["message"]

	p_ctx_map.append(msg_map)

	# FUNCTION_CALL
	if "function_call" in msg_map.keys():
		func_call_map = msg_map["function_call"]

		if p_verbose_bool:
			print(f"GPT function_call - {func_call_map['name']}, args: {func_call_map['arguments']}")
		
		return None, func_call_map, p_ctx_map

	# GENERATED TEXT
	else:
		
		generated_text_str = msg_map["content"]
		
		if p_verbose_bool:
			print(f"\n\nanswer: {generated_text_str}")

		return generated_text_str, None, p_ctx_map
	
	#----------------------

#--------------------------------------------
def model_request(p_prompt_str,
	p_api_key_str,
	p_model_str="gpt-4",
	p_functions_meta_map=None,
	p_verbose_bool=False):

	if p_verbose_bool:
		print(">>>>>>>>>>>>>>>> model request...")

	ctx_lst = [] 
	
	if p_verbose_bool:
		print(f"""
		
		prompt: 
			{p_prompt_str}
		
		""")
	
	gpt_msg_map = {
		"role": "user",
		"content": p_prompt_str
	}
	ctx_lst.append(gpt_msg_map)

	data_map = {
		"model": p_model_str, # "gpt-3.5-turbo",
		"messages": [
			gpt_msg_map
		],
		"temperature": 0.7
	}

	# if function signatures metadata is supplied, pass it to the model
	# for it to be able to use it in the completion.
	if not p_functions_meta_map == None:

		# definitions of functions meant for the GPT model
		functions_gpt_defs_lst = [v["gpt_def_map"] for _, v in p_functions_meta_map.items()]

		data_map["functions"] = functions_gpt_defs_lst

	'''
	payload = {
		"prompt": text_prompt,
		"max_tokens": 100,
	}
	'''
	
	r_map = make_request(data_map, p_api_key_str)



	#----------------------
	# MSG

	reply_gpt_msg_map = r_map["choices"][0]["message"]

	ctx_lst.append(reply_gpt_msg_map)

	# FUNCTION_CALL
	if "function_call" in reply_gpt_msg_map.keys():
		func_call_map = reply_gpt_msg_map["function_call"]

		if p_verbose_bool:
			print(f"GPT function_call - {func_call_map['name']}, args: {func_call_map['arguments']}")
		
		return None, func_call_map, ctx_lst

	# GENERATED TEXT
	else:
		
		generated_text_str = reply_gpt_msg_map["content"]
		
		if p_verbose_bool:
			print(f"\n\nanswer: {generated_text_str}")

		return generated_text_str, None, ctx_lst
	
	#----------------------

#--------------------------------------------
def make_request(p_data_map,
	p_api_key_str,
	p_verbose_bool=False):

	# Set the API endpoint for GPT-4 (if available)
	# api_endpoint = "https://api.openai.com/v1/engines/gpt-4/completions"
	api_endpoint = "https://api.openai.com/v1/chat/completions"

	headers = {
		"Content-Type": "application/json",
		"Authorization": f"Bearer {p_api_key_str}",
	}

	# HTTP_POST
	response = requests.post(api_endpoint, json=p_data_map, headers=headers)

	if response.status_code == 200:

		r_map = response.json()

		if p_verbose_bool:
			print("GPT response:")
			print(r_map)

		#----------------------
		# METADATA
		query_run_unix_time_int = r_map["created"]
		model_used_str = r_map["model"]

		#----------------------
		# METRICS
		total_tokens_consumed_int = r_map["usage"]["total_tokens"]
		prompt_tokens_int         = r_map["usage"]["prompt_tokens"]
		completion_tokens_int     = r_map["usage"]["completion_tokens"]


		return r_map
	
	else:
		print("Error:", response.status_code, response.text)

		with sentry_sdk.push_scope() as scope:
			scope.set_extra("status_code", response.status_code)
			scope.set_extra("response_text", response.text)
	
			raise Exception("Error:", response.status_code, response.text)
		

#--------------------------------------------
def dalle3_request(p_prompt_str,
	p_api_key_str):

	url_str = "https://api.openai.com/v1/engines/dall-e-3/completions"

	data_map = {
		"prompt": p_prompt_str,
		"max_tokens": 256,
	}

	headers = {
		"Content-Type": "application/json",
		"Authorization": f"Bearer {p_api_key_str}",
	}

	resp = requests.post(url_str, json=data_map, headers=headers)

	if resp.status_code == 200:
		image_url_str = resp.json()["choices"][0]["text"]
		return image_url_str

		# Optionally, you can download the image using additional code
		# For example, using the requests module:
		# image_response = requests.get(image_url)
		# with open("generated_image.png", "wb") as image_file:
		#     image_file.write(image_response.content)

	else:
		print("Error:", resp.status_code, resp.text)


# GloFlow application and media management/publishing platform
# Copyright (C) 2024 Ivan Trajkovic
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

# IMPORTANT!! - this module is run by GloFlow platform, not directly by user.
#       its run by the GF golang code, and its output thats prefixed 
#       with "GF_OUT:" is parsed by the GF golang code.

import os
import json
import argparse



#----------------------------------------------
def run():

    print("\nGF_IMAGES_CLASSIFY >>>>> PY \n")

    #--------------------
    # INPUT
    
    parser = argparse.ArgumentParser(description='image GF IDs to classify')
    parser.add_argument('-images_ids', type=str, help='list of image GF IDs to classify, comma-separated')
    args = parser.parse_args()

    sentry_dsn_str = os.getenv('SENTRY_DSN')
    sentry_env_str = os.getenv('SENTRY_ENV')

    #--------------------
    # SENTRY

    if not sentry_dsn_str == None:
        print("sentry enabled...")
        import sentry_sdk
        sentry_sdk.init(
            dns=sentry_dsn_str,
            environment=sentry_env_str,
            traces_sample_rate=1.0
        )

    #--------------------
    images_ids_lst = args.images_ids.split(',')


    print("images classify...")
    print(f"image id's: {images_ids_lst}")



    classes_lst = ['cat', 'dog', 'bird', 'fish']

    output_str = f"GF_OUT:{json.dumps({'classes_lst': classes_lst})}" 
    print(output_str)

#----------------------------------------------


if __name__ == "__main__":
    run()


# GloFlow application and media management/publishing platform
# Copyright (C) 2021 Ivan Trajkovic
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

import boto3
from botocore.exceptions import ClientError

#-------------------------------------------------------------
def describe_instances_by_tags(p_tags_map,
	p_region_str="us-east-1"):
	
	client = boto3.client("ec2", region_name=p_region_str)

	custom_filter = []
	for k, v in p_tags_map.items():
		custom_filter.append({
			"Name":   f"tag:{k}", 
			"Values": [v]
		})
	
	response = client.describe_instances(Filters=custom_filter)

	aws_instances_lst = []
	for i in response["Reservations"][0]["Instances"]:
		aws_instances_lst.append(i)

	return aws_instances_lst

#-------------------------------------------------------------
# SECURITY_GROUPS
#-------------------------------------------------------------
def get_secgr_id_by_name(p_name_str, p_region_str="us-east-1"):
	assert(isinstance(p_name_str, str))

	client = boto3.client("ec2", region_name=p_region_str)
	
	response = client.describe_security_groups(
		Filters=[{"Name": "group-name", "Values": [p_name_str]}]
	)

	if response["SecurityGroups"]:
		id_str = response["SecurityGroups"][0]["GroupId"]
		return id_str
	else:
		print(f"Security group '{p_name_str}' not found.")
		return None
	
#---------------------------------------------------------------------------------
def add_secgr_ingress_ip(p_ip_cidr_str,
	p_port_int,
	p_sec_gr_id_str,
	p_region_str="us-east-1"):
	assert(isinstance(p_ip_cidr_str, str))
	assert(isinstance(p_port_int, int))
	assert(isinstance(p_sec_gr_id_str, str))

	try:
		client = boto3.client("ec2", region_name=p_region_str)
		
		response = client.authorize_security_group_ingress(
			GroupId=p_sec_gr_id_str,
			IpPermissions=[
				{
					'IpProtocol': 'tcp',
					'FromPort': p_port_int,
					'ToPort': p_port_int,
					'IpRanges': [{'CidrIp': p_ip_cidr_str}]
				}
			]
		)
		
		if response['Return'] is not True:
			raise Exception(f"Ingress rule addition failed: {response}")

	except ClientError as e:
		raise Exception(f"Failed to add ingress rule: {e}")
	
#---------------------------------------------------------------------------------
def delete_secgr_ingress_ip(p_ip_cidr_str,
	p_port_int,
	p_sec_gr_id_str,
	p_region_str="us-east-1"):
	assert(isinstance(p_ip_cidr_str, str))
	assert(isinstance(p_port_int, int))
	assert(isinstance(p_sec_gr_id_str, str))

	try:
		client = boto3.client("ec2", region_name=p_region_str)

		response = client.revoke_security_group_ingress(
			GroupId=p_sec_gr_id_str,
			IpPermissions=[
				{
					'IpProtocol': 'tcp',
					'FromPort': p_port_int,
					'ToPort': p_port_int,
					'IpRanges': [{'CidrIp': p_ip_cidr_str}]
				}
			]
		)
		
		if response['Return'] is not True:
			raise Exception(f"Ingress rule revoking failed: {response}")
		
	except ClientError as e:
		raise Exception(f"Failed to revoke ingress rule: {e}")

#---------------------------------------------------------------------------------
def check_secgr_ingress_ip(p_ip_cidr_str,
	p_port_int,
	p_sec_gr_id_str,
	p_region_str="us-east-1"):
	assert(isinstance(p_ip_cidr_str, str))
	assert(isinstance(p_port_int, int))
	assert(isinstance(p_sec_gr_id_str, str))
	
	try:
		client   = boto3.client("ec2", region_name=p_region_str)
		response = client.describe_security_groups(GroupIds=[p_sec_gr_id_str])

		for permission in response['SecurityGroups'][0]['IpPermissions']:
			if (permission['FromPort'] == p_port_int and
				permission['ToPort'] == p_port_int and
				permission['IpProtocol'] == 'tcp'):
				for ip_range in permission['IpRanges']:
					if ip_range['CidrIp'] == p_ip_cidr_str:
						return True
		return False

	except ClientError as e:
		raise Exception(f"Failed to check ingress rule: {e}")

#---------------------------------------------------------------------------------
# check if the IP/port is allowed in the security group, and if its not add it

def ensure_secgr_ingress_ip(p_ip_cidr_str,
	p_port_int,
	p_sec_gr_id_str,
	p_region_str="us-east-1"):
	assert(isinstance(p_ip_cidr_str, str))
	assert(isinstance(p_port_int, int))
	assert(isinstance(p_sec_gr_id_str, str))

	if not check_secgr_ingress_ip(p_ip_cidr_str, p_port_int, p_sec_gr_id_str, p_region_str):
		add_secgr_ingress_ip(p_ip_cidr_str, p_port_int, p_sec_gr_id_str, p_region_str)


# GloFlow application and media management/publishing platform
# Copyright (C) 2019 Ivan Trajkovic
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

import os, sys
modd_str = os.path.abspath(os.path.dirname(__file__)) # module dir

import json
import subprocess
import base64
from colored import fg, bg, attr
import delegator

sys.path.append("%s/../../gf_core"%(modd_str))
import gf_core_cli

#---------------------------------------------------
def cont_is_running(p_cont_name_str,
	p_log_fun,
	p_exit_on_fail_bool = True,
	p_docker_sudo_bool  = True):

	sudo_str = ""
	if p_docker_sudo_bool:
		sudo_str = "sudo"

	stdout_str, stderr_str, exit_code_int = gf_core_cli.run("%s docker ps -a | grep %s"%(sudo_str, p_cont_name_str))

	if not stderr_str == "":
		print(stderr_str)
		
	# IMPORTANT!! - failure to reach Dcoerk daemon should always exit. its not a expected failure.
	if "Cannot connect to the Docker daemon" in stderr_str:
		exit(1)

	if stdout_str == "":
		print("CONTAINER NOT RUNNING -----------------------")
		return False
	else:
		print("CONTAINER RUNNING -----------------------")
		return True

#---------------------------------------------------
def cont_is_running_remote(p_cont_name_str,
	p_fab_conn,
	p_log_fun,
	p_exit_on_fail_bool = True,
	p_docker_sudo_bool  = True):

	print(f"{fg('yellow')}CHECK IF CONTAINER IS RUNNING{attr(0)} - {fg('green')}%s{attr(0)}"%(p_cont_name_str))
	sudo_str = ""
	if p_docker_sudo_bool:
		sudo_str = "sudo"

	c_str = "%s docker ps -a | grep %s"%(sudo_str, p_cont_name_str)
	print(c_str)
	r = p_fab_conn.run(c_str, warn=True)
	if not r.stdout == "": print(r.stdout)
	if not r.stderr == "": print(r.stderr)



	# IMPORTANT!! - failure to reach Dcoerk daemon should always exit. its not a expected failure.
	if "Cannot connect to the Docker daemon" in r.stderr:
		print("cannot connect to Docker daemon")
		exit(1)
	
	# # IMPORTANT!! - if command returns a non-zero exit code in some environments (CI) we
    # #               want to fail with that a non-zero exit code - this way CI will flag builds as failed.
	# #               in other scenarious its acceptable for this command to fail, and we want the caller
	# #               to keep executing.
	# if not r.return_code == 0:
	# 	if p_exit_on_fail_bool:
	# 		exit(r.return_code)

	if r.stdout == "":
		print(f"{fg('yellow')}CONTAINER NOT RUNNING{attr(0)}")
		return False
	else:
		print(f"{fg('green')}CONTAINER RUNNING{attr(0)}")
		return True

#-------------------------------------------------------------
# RUN
def run(p_full_image_name_str,
	p_log_fun,
	p_container_name_str = None,
	p_ports_map          = None,
	p_volumes_map        = None,
	p_hostname_str       = None,
	p_host_network_bool  = False,
	p_detached_bool      = True,
	p_exit_on_fail_bool  = False,
	p_docker_sudo_bool   = True):
	assert isinstance(p_full_image_name_str, str)

	print("")
	print("RUNNING DOCKER CONTAINER - %s"%(p_full_image_name_str))

	cmd_lst = []
	if p_docker_sudo_bool:
		cmd_lst.append("sudo")

	cmd_lst.extend([
		"docker run",
		"--restart=always",
	])

	# CONTAINER_NAME
	if not p_container_name_str == None:
		cmd_lst.append("--name %s"%(p_container_name_str))

	# PORTS
	if not p_ports_map == None:
		for host_port_str, container_port_str in p_ports_map.items():
			# IMPORTANT!! - "-p" publish a container's port or a range of ports to the host.
			cmd_lst.append("-p %s:%s"%(host_port_str, container_port_str))

	# VOLUMES
	if not p_volumes_map == None:
		for host_dir_str, container_dir_str in p_volumes_map.items():
			# IMPORTANT!! - "-v" - mount a host directory into a particular directory path in the
			#                      container filesystem.
			cmd_lst.append("-v %s:%s"%(host_dir_str, container_dir_str))

	# HOSTNAME
	if not p_hostname_str == None:
		cmd_lst.append("-h %s"%(p_hostname_str))

	# HOST_NETWORK
	if p_host_network_bool:
		cmd_lst.append("--net=host")

	# DETACHED
	if p_detached_bool:
		cmd_lst.append("-d")


	# IMAGE_NAME
	cmd_lst.append(p_full_image_name_str)


	c_str = " ".join(cmd_lst)
	p_log_fun("INFO", " - %s"%(c_str))

	stdout_str, stderr_str, exit_code_int = gf_core_cli.run(c_str)

	if not stderr_str == "":
		print(stderr_str)

	# IMPORTANT!! - failure to reach Dcoerk daemon should always exit. its not a expected failure.
	if "Cannot connect to the Docker daemon" in stderr_str:
		exit(1)

	# IMPORTANT!! - if command returns a non-zero exit code in some environments (CI) we
    #               want to fail with that a non-zero exit code - this way CI will flag builds as failed.
	#               in other scenarious its acceptable for this command to fail, and we want the caller
	#               to keep executing.
	if not exit_code_int == 0:
		if p_exit_on_fail_bool:
			exit(exit_code_int)

	# CONTAINER_ID
	container_id_str = stdout_str.strip()
	assert len(container_id_str) == 64

	return container_id_str

#-------------------------------------------------------------
# REMOVE
def remove_by_name(p_container_name_str,
	p_log_fun,
	p_exit_on_fail_bool = False,
	p_docker_sudo_bool  = True):

	sudo_str = ""
	if p_docker_sudo_bool:
		sudo_str = "sudo"

	cmd_str = "%s docker rm -f `%s docker ps -a | grep %s | awk '{print $1}'`"%(sudo_str, sudo_str, p_container_name_str)
	stdout_str, stderr_str, exit_code_int = gf_core_cli.run(cmd_str)

	if not stderr_str == "":
		print(stderr_str)
		
	# IMPORTANT!! - failure to reach Dcoerk daemon should always exit. its not a expected failure.
	if "Cannot connect to the Docker daemon" in stderr_str:
		exit(1)

	# IMPORTANT!! - if command returns a non-zero exit code in some environments (CI) we
    #               want to fail with that a non-zero exit code - this way CI will flag builds as failed.
	#               in other scenarious its acceptable for this command to fail, and we want the caller
	#               to keep executing.
	if not exit_code_int == 0:
		if p_exit_on_fail_bool:
			exit(exit_code_int)

#-------------------------------------------------------------
def remove_by_name_remote(p_container_name_str,
	p_fab_conn,
	p_exit_on_fail_bool = True,
	p_docker_sudo_bool  = True):

	sudo_str = ""
	if p_docker_sudo_bool:
		sudo_str = "sudo"

	cmd_str       = "%s docker rm -f `%s docker ps -a | grep %s | awk '{print $1}'`"%(sudo_str, sudo_str, p_container_name_str)
	r             = p_fab_conn.run(cmd_str)
	exit_code_int = r.return_code

	stdout_and_stderr_str = r.stdout+r.stderr
	
	# IMPORTANT!! - failure to reach Dcoerk daemon should always exit. its not a expected failure.
	if "Cannot connect to the Docker daemon" in r.stdout:
		exit(1)

	# IMPORTANT!! - if command returns a non-zero exit code in some environments (CI) we
    #               want to fail with that a non-zero exit code - this way CI will flag builds as failed.
	#               in other scenarious its acceptable for this command to fail, and we want the caller
	#               to keep executing.
	if not exit_code_int == 0:
		if p_exit_on_fail_bool:
			exit(exit_code_int)

#-------------------------------------------------------------
# PULL_IMAGE
def pull_remote(p_cont_image_name_str,
	p_fab_conn,
	p_log_fun,
	p_docker_sudo_bool  = False):
	p_log_fun("FUN_ENTER", "gf_os_docker.pull_image()")

	sudo_str = ""
	if p_docker_sudo_bool:
		sudo_str = "sudo"

	c_str = "%s docker pull %s"%(sudo_str, p_cont_image_name_str)
	p_fab_conn.run(c_str)

#-------------------------------------------------------------
# PUSH
def push(p_image_full_name_str,
	p_docker_user_str,
	p_docker_pass_str,
	p_log_fun,
	p_host_str          = None,
	p_exit_on_fail_bool = False,
	p_docker_sudo_bool  = False):
	p_log_fun("FUN_ENTER", "gf_os_docker.push()")
	p_log_fun("INFO",      f"image_full_name - {p_image_full_name_str}")
	assert isinstance(p_docker_user_str, str)

	#------------------
	# LOGIN

	login(p_docker_user_str,
		p_docker_pass_str,
		p_host_str          = p_host_str,
		p_exit_on_fail_bool = p_exit_on_fail_bool,
		p_docker_sudo_bool  = p_docker_sudo_bool)

	#------------------
	cmd_lst = []
	if p_docker_sudo_bool:
		cmd_lst.append("sudo")

	cmd_lst.extend([
		"docker push",
		p_image_full_name_str
	])

	c_str = " ".join(cmd_lst)
	p_log_fun("INFO", " - %s"%(c_str))

	stdout_str, stderr_str, exit_code_int = gf_core_cli.run(c_str)

	if not stderr_str == "":
		print(stderr_str)
		
	# IMPORTANT!! - failure to reach Dcoerk daemon should always exit. its not a expected failure.
	if "Cannot connect to the Docker daemon" in stderr_str:
		exit(1)

	# IMPORTANT!! - if command returns a non-zero exit code in some environments (CI) we
    #               want to fail with that a non-zero exit code - this way CI will flag builds as failed.
	#               in other scenarious its acceptable for this command to fail, and we want the caller
	#               to keep executing.
	if not exit_code_int == 0:
		if p_exit_on_fail_bool:
			exit(exit_code_int)

	#------------------
	# DOCKER_LOGOUT
	cmd_lst = []
	if p_docker_sudo_bool:
		cmd_lst.append("sudo")
	cmd_lst.append("docker logout")
	stdout_str, _, _ = gf_core_cli.run(" ".join(cmd_lst))
	print(stdout_str)

	#------------------

#-------------------------------------------------------------
# BUILD_IMAGE
def build_image(p_image_names_lst,
	p_dockerfile_path_str,
	p_log_fun,
	p_build_args_map    = None,
	p_exit_on_fail_bool = False,
	p_docker_sudo_bool  = False,
	p_pull_base_bool    = False,
	p_ctx_dir_str = None):
	
	# p_log_fun("FUN_ENTER", "gf_os_docker.build_image()")
	
	assert isinstance(p_image_names_lst, list)
	print(p_dockerfile_path_str)
	assert os.path.isfile(p_dockerfile_path_str)
	assert "Dockerfile" in os.path.basename(p_dockerfile_path_str)

	# full_image_name_str  = "%s/%s:%s"%(p_user_name_str, p_image_name_str, p_image_tag_str)

	# CONTEXT_DIR
	if p_ctx_dir_str == None:
		context_dir_path_str = os.path.dirname(p_dockerfile_path_str)
	else:
		context_dir_path_str = p_ctx_dir_str

	p_log_fun("INFO", "====================+++++++++++++++=====================")
	p_log_fun("INFO", "                 BUILDING DOCKER IMAGE")
	p_log_fun("INFO", "image_names - %s"%(p_image_names_lst))
	p_log_fun("INFO", "Dockerfile  - %s"%(p_dockerfile_path_str))
	p_log_fun("INFO", "====================+++++++++++++++=====================")

	cmd_lst = []

	# RUN_WITH_SUDO - Docker deamon runs as root, and so for docker client to be able to connect to it
	#                 without any custom config the client needs to be run with "sudo".
	#                 if some config is in place to avoid this, set p_docker_sudo_bool to False.
	if p_docker_sudo_bool:
		cmd_lst.append("sudo")
		
	cmd_lst.extend([
		"docker", "build",
		"-f", p_dockerfile_path_str,
	])

	# pull base image
	if p_pull_base_bool:
		cmd_lst.append("--pull")

	# TAGS - there can be multiple tags for the same image
	for n in p_image_names_lst:
		cmd_lst.append("--tag=%s"%(n))

	# BUILD_ARGS
	if not p_build_args_map == None:
		for k, v in p_build_args_map.items():
			cmd_lst.extend(["--build-arg", "%s=%s"%(k, v)])

	# CONTEXT_DIR
	cmd_lst.append(context_dir_path_str)

	c_str = " ".join(cmd_lst)
	p_log_fun("INFO", c_str)

	# change to the dir where the Dockerfile is located, for the 'docker'
	# tool to have the proper context
	old_cwd = os.getcwd()
	os.chdir(context_dir_path_str)
	
	#---------------------------------------------------
	def get_image_id_from_line(p_stdout_line_str):
		p_lst = p_stdout_line_str.split(" ")

		assert len(p_lst) == 3
		image_id_str = p_lst[2]

		# IMPORTANT!! - check that this is a valid 12 char Docker ID
		assert len(image_id_str) == 12
		return image_id_str

	#---------------------------------------------------

	# IMPORTANT!! - outputs stdout/stderr of the command shell run in realtime via
	#               threads that are spawned to print to screen.
	p = gf_core_cli.run__view_realtime(cmd_lst, {},
        "gf_build_cont", "green")
	p.wait()

	exit_code_int = p.returncode


	# IMPORTANT!! - if command returns a non-zero exit code in some environments (CI) we
    #               want to fail with that a non-zero exit code - this way CI will flag builds as failed.
	#               in other scenarious its acceptable for this command to fail, and we want the caller
	#               to keep executing.
	if not exit_code_int == 0:
		print("build failed - non-zero exit code returned...")
		if p_exit_on_fail_bool:
			exit(exit_code_int)

	# change back to old dir
	os.chdir(old_cwd)

#-------------------------------------------------------------
# LOGIN
def login(p_docker_user_str,
	p_docker_pass_str,
	p_host_str          = None,
	p_exit_on_fail_bool = True,
	p_docker_sudo_bool  = False):
	assert isinstance(p_docker_user_str, str)
	assert isinstance(p_docker_pass_str, str)



	cmd_lst = []
	if p_docker_sudo_bool:
		cmd_lst.append("sudo")
		
	cmd_lst.extend([
		"docker", "login",
		"-u", p_docker_user_str,
		"--password-stdin"
	])

	# HOST
	if not p_host_str == None:
		cmd_lst.append(p_host_str)

	cmd_str = " ".join(cmd_lst)
	print(cmd_str)
	
	process = subprocess.Popen(cmd_str,
		shell=True,
		stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
	
	# STDIN_WRITE
	process.stdin.write(bytes(p_docker_pass_str.encode())) # write password to stdin of "docker login" command
	
	stdout, stderr = process.communicate() # wait for command completion
	stdout_str = stdout.decode()
	stderr_str = stderr.decode()
	print(stdout_str)

	if not stderr_str == "":
		print(stderr_str)
		
	# IMPORTANT!! - failure to reach Dcoerk daemon should always exit. its not a expected failure.
	if "Cannot connect to the Docker daemon" in stderr_str:
		exit(1)

	# IMPORTANT!! - if command returns a non-zero exit code in some environments (CI) we
    #               want to fail with that a non-zero exit code - this way CI will flag builds as failed.
	#               in other scenarious its acceptable for this command to fail, and we want the caller
	#               to keep executing.
	if p_exit_on_fail_bool:
		if not process.returncode == 0:
			exit()

#---------------------------------------------------
# LOGIN__REMOTE
def login__remote(p_docker_user_str,
	p_docker_pass_str,
	p_fab_conn,
	p_log_fun):
	p_log_fun("FUN_ENTER", "gf_os_docker.login__remote()")
	assert isinstance(p_docker_user_str, str)
	assert isinstance(p_docker_pass_str, str)

	import fabric
	
	#---------------------------
	# UPLOAD_PASS_FILE

	pass_f_str = "tmp_file"
	f = open(pass_f_str, "w")
	f.write(p_docker_pass_str)
	f.close()
	
	fabric.api.put(pass_f_str) # upload password file
	
	#---------------------------
	# IMPORTANT!! - specify pasword from stdin so that it doesnt show up
	#               as a part of the final command (in logs)

	c_str = "cat %s | sudo docker login -u %s --password-stdin"%(pass_f_str, p_docker_user_str)
	p_fab_conn.run(c_str)
	# fabric.api.run(c_str)

	#---------------------------
	p_fab_conn.run("rm %s"%(pass_f_str))
	delegator.run("rm %s"%(pass_f_str)) # clean local tmp_file that holds the dockerhub password

#---------------------------------------------------
# LOGIN__REMOTE_FROM_FILE
def login__remote_from_file(p_docker_user_str,
	p_docker_pass_str,
	p_fab_conn,
	p_log_fun,
	p_ecr_bool           = False,
	p_aws_account_id_str = None,
	p_region_str         = "us-east-1"):
	p_log_fun("FUN_ENTER", "gf_os_docker.login__remote_from_file()")
	assert isinstance(p_docker_user_str, str)
	assert isinstance(p_docker_pass_str, str)

	#---------------------------
	# UPLOAD_PASS_FILE

	pass_f_str = "tmp_file"
	f = open(pass_f_str, "w")
	f.write(p_docker_pass_str)
	f.close()

	p_fab_conn.put(pass_f_str) # upload password file

	#---------------------------
	# IMPORTANT!! - specify pasword from stdin so that it doesnt show up
	#               as a part of the final command (in logs)

	if p_ecr_bool:
		assert not p_aws_account_id_str == None
		ecr_acc_str = f"{p_aws_account_id_str}.dkr.ecr.{p_region_str}.amazonaws.com"
		p_fab_conn.run(f"cat {pass_f_str} | sudo docker login -u {p_docker_user_str} --password-stdin {ecr_acc_str}")
	else:
		p_fab_conn.run(f"cat {pass_f_str} | sudo docker login -u {p_docker_user_str} --password-stdin")
	
	#---------------------------
	p_fab_conn.run("rm %s"%(pass_f_str))
	delegator.run("rm %s"%(pass_f_str)) # clean local tmp_file that holds the dockerhub password

#---------------------------------------------------
# CLEAN_STOP__CONTAINERS
def clean_stop__containers(p_cont_image_name_str,
	p_fab_conn,
	p_log_fun):
	p_log_fun("FUN_ENTER", "gf_os_docker.clean_stop__containers()")

	#--------------------
	# STOP_CURRENT_CONTAINERS
	image_ids_str = p_fab_conn.run("sudo docker ps -a | grep %s | awk '{print $1}'"%(p_cont_image_name_str))
	print("image_ids_str - %s"%(image_ids_str))

	if not image_ids_str == "":
		print("    >>  image already running - %s"%(p_cont_image_name_str))
		print("    >>  stoping containers    - %s"%(p_cont_image_name_str))

		for l in image_ids_str.split("\n"):
			image_id_str = l
			p_fab_conn.run("sudo docker stop %s"%(image_id_str)) # stop first
			p_fab_conn.run("sudo docker rm %s"%(image_id_str))   # remove, to not conflict with new ones

	#--------------------

#---------------------------------------------------
def install_base_docker(p_fab_conn,
	p_log_fun,
	p_os_type_str = "ubuntu"):
	p_log_fun("FUN_ENTER", "gf_os_docker.install_base_docker()")
	assert p_os_type_str == "ubuntu" or p_os_type_str == "amazon_linux_2"
	
	if p_os_type_str == "ubuntu":
		p_fab_conn.run("sudo apt-get clean")
		p_fab_conn.run("sudo apt-get update")
		p_fab_conn.run("sudo apt-get install -y apt-transport-https ca-certificates curl software-properties-common")
		p_fab_conn.run("sudo curl -fsSL https://download.docker.com/linux/ubuntu/gpg | sudo apt-key add -")
		

		# last one validated for Kubernetes installations
		# docker_version_str = "17.06.0~ce-0~ubuntu"

		##FIX!! - hardcoding to "zesty" ubuntu version (17.04) because in 17.10 at the moment (dec 10 2017) there is no docker-ce package
		##        so Im hardcdoing 17.04 just for the moment so that the compatible docker-ce package is used
		##p_fab_api.run('sudo add-apt-repository "deb [arch=amd64] https://download.docker.com/linux/ubuntu $(lsb_release -cs) stable"')
		#p_fab_api.run('sudo add-apt-repository "deb [arch=amd64] https://download.docker.com/linux/ubuntu zesty stable"')

		p_fab_conn.run("sudo apt-get update")
		p_fab_conn.run("sudo apt-get install -y \
			apt-transport-https \
			ca-certificates \
			curl \
			gnupg-agent \
			software-properties-common")
		p_fab_conn.run("curl -fsSL https://download.docker.com/linux/ubuntu/gpg | sudo apt-key add -")
		p_fab_conn.run('sudo add-apt-repository \
			"deb [arch=amd64] https://download.docker.com/linux/ubuntu \
			$(lsb_release -cs) \
			stable"')
		p_fab_conn.run("sudo apt-get update")
		p_fab_conn.run("sudo apt-get install -y docker-ce docker-ce-cli containerd.io")

	elif p_os_type_str == "amazon_linux_2":

		p_fab_conn.run("sudo yum -y install docker")
		p_fab_conn.run("sudo systemctl start docker")


#---------------------------------------------------
def dockerhub__get_auth_config_json(p_docker_user_str,
	p_docker_pass_str,
	p_log_fun):
	p_log_fun("FUN_ENTER", "gf_os_docker.dockerhub__get_auth_config_json()")

	auth_str             = base64.b64encode("%s:%s"%(p_docker_user_str, p_docker_pass_str))
	auth_config_map      = {"auths": {"https://index.docker.io/v1/": {"auth": auth_str}}}
	auth_config_json_str = json.dumps(auth_config_map)
	return auth_config_json_str


# GloFlow application and media management/publishing platform
# Copyright (C) 2020 Ivan Trajkovic
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

import os, sys
modd_str = os.path.abspath(os.path.dirname(__file__)) # module dir

from colored import fg, bg, attr

sys.path.append("%s/../utils"%(modd_str))
import gf_core_cli

import gf_test__sqs_process
import gf_test__master_get_block

#--------------------------------------------------
def run_go(p_test_ci_bool):

	aws_region_str = "us-east-1"
	py_plugins_base_dir_path_str = f"{modd_str}/../plugins"
	
	# CI
	if p_test_ci_bool:
		
		p__worker_inspector = init_ci(py_plugins_base_dir_path_str,
			aws_region_str)

	#--------------------------------------------------
	def run__tests():

		cmd_lst = [
			"go", "test", "-v"
		]
		p = gf_core_cli.run__view_realtime(cmd_lst, {
				"GF_GETH_HOST":       os.environ["GF_GETH_HOST"],
				"GF_SENTRY_ENDPOINT": os.environ["GF_SENTRY_ENDPOINT"],

				# p_py_plugins_base_dir_path_str
				"GF_PY_PLUGINS_BASE_DIR_PATH": os.environ["GF_PY_PLUGINS_BASE_DIR"]
			},
			"gf_eth_monitor__worker_inspector", "cyan")
	
	#--------------------------------------------------
	run__tests()

#--------------------------------------------------
def run_py(p_test_ci_bool):
	
	# assert "AWS_ACCESS_KEY_ID" in os.environ.keys()
	# assert "AWS_SECRET_ACCESS_KEY" in os.environ.keys()
	
	assert "GF_AWS_SQS_QUEUE" in os.environ.keys()
	assert "GF_SENTRY_ENDPOINT" in os.environ.keys()


	aws_region_str               = "us-east-1"
	py_plugins_base_dir_path_str = f"{modd_str}/../plugins"

	# CI
	if p_test_ci_bool:
		
		p__worker_inspector = init_ci(py_plugins_base_dir_path_str,
			aws_region_str)

	# LOCAL
	else:

		test_gf_geth_path_str = f"{modd_str}/../../../gloflow_go-ethereum/build/bin/geth"
		data_dir_path_str     = f"{modd_str}/test_data_geth"
		assert os.path.isfile(test_gf_geth_path_str)
		assert os.path.isdir(py_plugins_base_dir_path_str)

		p__geth, p__worker_inspector = init_local(test_gf_geth_path_str,
			data_dir_path_str,
			py_plugins_base_dir_path_str,
			aws_region_str)


	#------------------------
	# TESTS
	gf_test__sqs_process.run(p_test_ci_bool, aws_region_str)
	gf_test__master_get_block.run(p_test_ci_bool, aws_region_str)

	#------------------------


	# CI
	if p_test_ci_bool:
		p__worker_inspector.terminate()

	# LOCAL
	else:
		p__geth.terminate()
		p__worker_inspector.terminate()

#--------------------------------------------------
def init(p_py_plugins_base_dir_path_str,
	p_aws_region_str):
	True

#--------------------------------------------------
def init_ci(p_py_plugins_base_dir_path_str,
	p_aws_region_str,
	p_worker_inspector_bool = True):

	# DB
	init_db()

	if p_worker_inspector_bool:
		#-------------------------------------------------------------
		def start_worker_inspector():
			cmd_lst = [
				f"{modd_str}/../../build/gf_eth_monitor_worker_inspector"
			]
			p = gf_core_cli.run__view_realtime(cmd_lst, {
					"GF_GETH_HOST":                os.environ["GF_GETH_HOST"],
					"GF_SENTRY_ENDPOINT":          os.environ["GF_SENTRY_ENDPOINT"],
					"GF_PY_PLUGINS_BASE_DIR_PATH": p_py_plugins_base_dir_path_str,
				},
				"gf_eth_monitor__worker_inspector", "yellow")

			return p

		#-------------------------------------------------------------

		p__worker_inspector = start_worker_inspector()
		return p__worker_inspector

#--------------------------------------------------
def init_local(p_geth_bin_path_str,
	p_geth_data_dir_path_str,
	p_py_plugins_base_dir_path_str,
	p_aws_region_str):
	assert os.path.isdir(p_py_plugins_base_dir_path_str)

	#-------------------------------------------------------------
	# START_GETH
	def start_geth():
	  
		print("%srunning GETH...%s"%(fg("yellow"), attr(0)))
		
		assert "AWS_ACCESS_KEY_ID" in os.environ.keys()
		assert "AWS_SECRET_ACCESS_KEY" in os.environ.keys()
		assert "GF_AWS_SQS_QUEUE" in os.environ.keys()
		
		cmd_lst = [
			p_geth_bin_path_str,
			f"--datadir={p_geth_data_dir_path_str}",
			"--syncmode=fast",
			"--http",
			"--maxpeers=10"
		]
		c_str = " ".join(cmd_lst)
		print(c_str)

		# # When shell=True the shell is the child process, and the commands are its children.
		# # So any SIGTERM or SIGKILL will kill the shell but not its child processes.
		# # The best way I can think of is to use shell=False, otherwise when you kill
		# # the parent shell process, it will leave a defunct shell process.
		# # CMD also has to be a list here, since its not being passed in as a string
		# # to the child shell.
		# geth_p = subprocess.Popen(cmd_lst, shell=False, stdout=subprocess.PIPE, bufsize=1,
		# 	env={
		# 		"AWS_REGION":            p_aws_region_str,
		# 		"AWS_ACCESS_KEY_ID":     os.environ["AWS_ACCESS_KEY_ID"],
		# 		"AWS_SECRET_ACCESS_KEY": os.environ["AWS_SECRET_ACCESS_KEY"],
		# 		"GF_AWS_SQS_QUEUE":      os.environ["GF_AWS_SQS_QUEUE"]
		# 	})
		# t = threading.Thread(target=gf_test_utils.read_process_stdout, args=(geth_p.stdout, "geth", "cyan"))
		# t.start()

		geth_p = gf_core_cli.run__view_realtime(cmd_lst, {
				"AWS_REGION":            p_aws_region_str,
				"AWS_ACCESS_KEY_ID":     os.environ["AWS_ACCESS_KEY_ID"],
				"AWS_SECRET_ACCESS_KEY": os.environ["AWS_SECRET_ACCESS_KEY"],
				"GF_AWS_SQS_QUEUE":      os.environ["GF_AWS_SQS_QUEUE"]
			},
			"geth", "cyan")

		return geth_p

	#-------------------------------------------------------------
	def start_worker_inspector():
		cmd_lst = [
			f"{modd_str}/../../build/gf_eth_monitor_worker_inspector"
		]

		# p = subprocess.Popen(cmd_lst, shell=False, stdout=subprocess.PIPE, bufsize=1,
		# 	env={
		# 		"GF_GETH_HOST":                "127.0.0.1",
		# 		"GF_SENTRY_ENDPOINT":          os.environ["GF_SENTRY_ENDPOINT"],
		# 		"GF_PY_PLUGINS_BASE_DIR_PATH": p_py_plugins_base_dir_path_str,
		# 	})
		# t = threading.Thread(target=gf_test_utils.read_process_stdout, args=(p.stdout, "gf_eth_monitor__worker_inspector", "yellow"))
		# t.start()

		gf_core_cli.run__view_realtime(cmd_lst, {
				"GF_GETH_HOST":                "127.0.0.1",
				"GF_SENTRY_ENDPOINT":          os.environ["GF_SENTRY_ENDPOINT"],
				"GF_PY_PLUGINS_BASE_DIR_PATH": p_py_plugins_base_dir_path_str,
			},
			"gf_eth_monitor__worker_inspector", "yellow")

		return p

	#-------------------------------------------------------------


	p = start_geth()
	p__worker_inspector = start_worker_inspector()

	import time
	time.sleep(10) # give time to Geth to initialize and send some events

	print("done sleeping")

	return p, p__worker_inspector

#--------------------------------------------------
def init_db():
	

	ec2_abi_str = '''[
		{
			"constant": true,
			"inputs": [],
			"name": "name",
			"outputs": [
				{
					"name": "",
					"type": "string"
				}
			],
			"payable": false,
			"stateMutability": "view",
			"type": "function"
		},
		{
			"constant": false,
			"inputs": [
				{
					"name": "_spender",
					"type": "address"
				},
				{
					"name": "_value",
					"type": "uint256"
				}
			],
			"name": "approve",
			"outputs": [
				{
					"name": "",
					"type": "bool"
				}
			],
			"payable": false,
			"stateMutability": "nonpayable",
			"type": "function"
		},
		{
			"constant": true,
			"inputs": [],
			"name": "totalSupply",
			"outputs": [
				{
					"name": "",
					"type": "uint256"
				}
			],
			"payable": false,
			"stateMutability": "view",
			"type": "function"
		},
		{
			"constant": false,
			"inputs": [
				{
					"name": "_from",
					"type": "address"
				},
				{
					"name": "_to",
					"type": "address"
				},
				{
					"name": "_value",
					"type": "uint256"
				}
			],
			"name": "transferFrom",
			"outputs": [
				{
					"name": "",
					"type": "bool"
				}
			],
			"payable": false,
			"stateMutability": "nonpayable",
			"type": "function"
		},
		{
			"constant": true,
			"inputs": [],
			"name": "decimals",
			"outputs": [
				{
					"name": "",
					"type": "uint8"
				}
			],
			"payable": false,
			"stateMutability": "view",
			"type": "function"
		},
		{
			"constant": true,
			"inputs": [
				{
					"name": "_owner",
					"type": "address"
				}
			],
			"name": "balanceOf",
			"outputs": [
				{
					"name": "balance",
					"type": "uint256"
				}
			],
			"payable": false,
			"stateMutability": "view",
			"type": "function"
		},
		{
			"constant": true,
			"inputs": [],
			"name": "symbol",
			"outputs": [
				{
					"name": "",
					"type": "string"
				}
			],
			"payable": false,
			"stateMutability": "view",
			"type": "function"
		},
		{
			"constant": false,
			"inputs": [
				{
					"name": "_to",
					"type": "address"
				},
				{
					"name": "_value",
					"type": "uint256"
				}
			],
			"name": "transfer",
			"outputs": [
				{
					"name": "",
					"type": "bool"
				}
			],
			"payable": false,
			"stateMutability": "nonpayable",
			"type": "function"
		},
		{
			"constant": true,
			"inputs": [
				{
					"name": "_owner",
					"type": "address"
				},
				{
					"name": "_spender",
					"type": "address"
				}
			],
			"name": "allowance",
			"outputs": [
				{
					"name": "",
					"type": "uint256"
				}
			],
			"payable": false,
			"stateMutability": "view",
			"type": "function"
		},
		{
			"payable": true,
			"stateMutability": "payable",
			"type": "fallback"
		},
		{
			"anonymous": false,
			"inputs": [
				{
					"indexed": true,
					"name": "owner",
					"type": "address"
				},
				{
					"indexed": true,
					"name": "spender",
					"type": "address"
				},
				{
					"indexed": false,
					"name": "value",
					"type": "uint256"
				}
			],
			"name": "Approval",
			"type": "event"
		},
		{
			"anonymous": false,
			"inputs": [
				{
					"indexed": true,
					"name": "from",
					"type": "address"
				},
				{
					"indexed": true,
					"name": "to",
					"type": "address"
				},
				{
					"indexed": false,
					"name": "value",
					"type": "uint256"
				}
			],
			"name": "Transfer",
			"type": "event"
		}
	]'''

	import json
	def_lst = json.loads(ec2_abi_str)
	erc20_abi_map = {
		"type_str": "erc20",
		"def_lst":  def_lst
	}



	import pymongo
	db_name_str   = "gf_eth_monitor"
	coll_name_str = "gf_eth_meta__contracts_abi"
	client = pymongo.MongoClient('localhost', 27017)
	client[db_name_str][coll_name_str].insert(erc20_abi_map)


# GloFlow application and media management/publishing platform
# Copyright (C) 2019 Ivan Trajkovic
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

import datetime
import pprint
import pandas as pd
import matplotlib.pyplot as plt 

#-------------------------------------------------------------
#called to find out how frequently to run the stat
def freq():
	return '1m'
#-------------------------------------------------------------
def run(p_mongo_client,
	p_log_fun,
	p_output_img_str = '../plots/stat__crawler_url_fetches__counts_by_day.png'):

	#-------------------------------------------------------------
	def query():
		coll    = p_mongo_client['prod_db']['data_symphony']
		results = coll.aggregate([
				{'$match':{
					't': 'gf_error'}},

				{"$project":{
					"creation_unix_time_f": 1,
					"type_str":             1,
					"service_name_str":     1,
					"subsystem_name_str":   1,
					"func_name_str":        1,
				}},
				{"$group":{

					#group by day
					#composite ID to group by - _id.date_str,_id.subsystem_name_str
					"_id":{
						"date_str":{
							"$dateToString": {
							    "format":"%Y-%m-%d",
							    "date"  :{

							    	#convert creation_unix_time_f from seconds to a Date object
							        "$add":[
							            datetime.datetime(1970,1,1), #new Date(0), 

							            #convert creation_unix_time_f seconds to millisecond timestamp first through
							            #multiplying the value by 1000.
							            {"$multiply": [1000, "$creation_unix_time_f"]}
							        ]
							    }
							},
						},
						'subsystem_name_str':'$subsystem_name_str'
			        },
			        'count_int':{'$sum':1}, #gf_error count for a particular day
			    }},

			    #sort all _id.date_str,_id.subsystem_name_str groups by the number of gf_errors in them,
			    #so that once they're grouped by day, subsystems are already sorted by their counts (number
			    #of gf_errors in them). 
			    {'$sort':{'count_int':-1}},

			    {'$group':{
			    	'_id':                  "$_id.date_str",
			    	"subsystems_errors_lst":{"$push":{
			    		"subsystem_name_str":"$_id.subsystem_name_str",
			    		"count_int":         "$count_int" #count for the individual domain in that day
			    	}},
			    	"count_int":{"$sum":"$count_int"}, #add the count for each domain on that date
			    }},

			    #IMPORTANT!! - for plotting, oldest records need to be first, timeseries plots go from left to right
			    {'$sort':{'_id':1}},
			],
			allowDiskUse=True)

		print('DONE - allowDiskUse')
		return results

		#print results.explain("executionStats")
		#print coll.explain("executionStats")
	#-------------------------------------------------------------
	
	results = query()
	df_lst  = []

	for r in results:
		day_str                     = r['_id']
		gf_error_counts_per_day_int = r['count_int']

		d = {
			'day':                    day_str,
			'gf_error_counts_per_day':gf_error_counts_per_day_int,
		}

		print('============= ---------------')

		for subsystem_errors_map in r['subsystems_errors_lst']:

			print(subsystem_errors_map)

			subsystem_name_str            = subsystem_errors_map['subsystem_name_str']
			subsystem_gf_errors_count_int = subsystem_errors_map['count_int']

			assert isinstance(subsystem_name_str,           basestring)
			assert isinstance(subsystem_gf_errors_count_int,int)

			d['subsys__%s'%(subsystem_name_str)] = subsystem_gf_errors_count_int
	
		df_lst.append(d)

	df = pd.DataFrame(df_lst)

	df.set_index("day",drop=True,inplace=True)
	print(df)

	#---------------------------------------------
	#PLOT

	fig = plt.figure(figsize=(30,10))

	#casting subject_alt_names_counts_lst to list() first because its a "multiprocessing.managers.ListProxy"
	#count_s = pd.Series(results_lst)
	#l = count_s.value_counts(sort=False)
	#l.sort_index(inplace=True)
	#l.plot.bar(figsize=(10,6))

	df.plot.line(figsize=(10,6),alpha=0.75) #,rot=0)
	
	plt.title("gf_errors's counts per day (per subsystem)",fontsize=18)
	plt.xlabel("day",                                      fontsize=14)
	plt.ylabel('number of gf_errors',                      fontsize=14)
	plt.xticks(size = 6)
	plt.axes().yaxis.grid() #horizontal-grid

	plt.savefig(p_output_img_str) #save figure to file
	#---------------------------------------------


# GloFlow application and media management/publishing platform
# Copyright (C) 2020 Ivan Trajkovic
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

import boto3

#--------------------------------------------------
def init_client(p_region_str):
    
    aws_client = boto3.client("secretsmanager",
        region_name = p_region_str)
    return aws_client

#--------------------------------------------------
def get_secret(p_secret_name_str,
    p_client):

    secret_value = p_client.get_secret_value(SecretId=p_secret_name_str)
    secret_value_str = secret_value["SecretString"]
    return secret_value_str

#--------------------------------------------------
def create_or_update_batch(p_secrets_map,
    p_region_str,
    p_aws_creds_map):
    assert isinstance(p_secrets_map, dict)
    assert isinstance(p_aws_creds_map, dict)

    aws_client = boto3.client("secretsmanager",
        aws_access_key_id     = p_aws_creds_map["AWS_ACCESS_KEY_ID"],
        aws_secret_access_key = p_aws_creds_map["AWS_SECRET_ACCESS_KEY"],
        region_name           = p_region_str)

    r = aws_client.list_secrets()
    existing_secrets_lst = [secret_map["Name"] for secret_map in r["SecretList"]]

    for secret_name_str, secret_value_str in p_secrets_map.items():

        # UPDATE - if the secret already exists
        if secret_name_str in existing_secrets_lst:
            aws_client.update_secret(SecretId = secret_name_str,
                SecretString = secret_value_str)

        # CREATE - create a secret if it doesnt exist yet
        else:
            aws_client.create_secret(Name = secret_name_str,
                SecretString = secret_value_str)


# GloFlow application and media management/publishing platform
# Copyright (C) 2022 Ivan Trajkovic
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

# IMPORTANT!! - this module is run by GloFlow platform, not directly by user.
#       its run by the GF golang code, and its output thats prefixed 
#       with "GF_OUT:" is parsed by the GF golang code.

import time

#----------------------------------------------
def main():


    print("GF_IMAGES PLUGINS >>>>")

    time.sleep(1)






#----------------------------------------------
main()


# GloFlow application and media management/publishing platform
# Copyright (C) 2019 Ivan Trajkovic
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

import os
import logging
import subprocess
import signal
import multiprocessing

import json
import delegator

#---------------------------------------------------
def run_cmd_in_os_proc(p_cmd_str, p_log_fun):
	
	#---------------------------------------------------
	def run_process():

		p_log_fun("INFO", "RUN CMD IN OS_PROCESS")
		p_log_fun("INFO", "p_cmd_str - %s"%(p_cmd_str))

		p = subprocess.Popen(p_cmd_str,
			shell   = True,
			stdout  = subprocess.PIPE,
			bufsize = 1)
		assert isinstance(p,subprocess.Popen)
		#---------------------------------------------------
		# IMPORTANT!! - workers cleanup on parent shutdown

		def handle_signal_terminate(p_signum, p_frame):
			p_log_fun("INFO", "+++ ++ -- SIGNAL SIGTERM RECEIVED - gf_images_main_service")

			p.terminate()

		import signal
		signal.signal(signal.SIGTERM, handle_signal_terminate)

		#---------------------------------------------------

		bin_str = os.path.basename(p_cmd_str.split(" ")[0])
		print(envoy.run("ps -e | grep %s"%(bin_str)).std_out)

		p.wait() # block this process and let the child run
	#---------------------------------------------------

	multiprocessing.log_to_stderr(logging.DEBUG)
	parent_conn, child_conn = multiprocessing.Pipe()
	process                 = multiprocessing.Process(target = run_process)
	process.start()

#---------------------------------------------------
# GET_SELF_IP

def get_self_ip():

	#---------------------------------------------------
	# IMPORTANT!! - this approach works most of the time.
	#               in some providers or networks DNS traffic might be routed from different exit routers
	#               (access points) than regular http traffic. 
	#               in thoise situations DNS nameserver that is pinged will return IP that is not appropriate.
	def dns_method():
		target_namespace_server_str = "ns1.google.com"

		# "-4" - force discovery of ipv4 address, since some ISP's by default return ipv6
		cmd_str = '''dig -4 TXT +short o-o.myaddr.l.google.com @%s | awk -F'"' '{ print $2}' '''%(target_namespace_server_str)
		print(cmd_str)

		r           = delegator.run(cmd_str)
		self_ip_str = r.out.strip()
		# assert len(self_ip_str.split(".")) == 4
		return self_ip_str

	#---------------------------------------------------
	def extern_service_ipinfoio_method():

		#---------------------------------------------------
		def get_remote():

			cmd_str = "curl http://ipinfo.io"
			print(cmd_str)
			
			r = delegator.run(cmd_str)

			self_ip_str = json.loads(r.out)["ip"]

			f=open("self_ip_cache.txt", "w")
			f.write(self_ip_str)


			return self_ip_str

		#---------------------------------------------------

		ip_str = get_remote()
		return ip_str
		
	#---------------------------------------------------
	def extern_service_ifconfigme_method():
		r = delegator.run("curl ifconfig.me")
		ip_str = r.out
		return ip_str

	#---------------------------------------------------
	def discover_ip():


		# IMPORTANT!! - in some networks in some countries many exit points are used dynamically for the same
		#               access point (mobile router); and each of these methods determines a different IP.
		#               so we're executing them all here and returning them all.
		dns__self_ip_str        = dns_method()
		ipinfo__self_ip_str     = extern_service_ipinfoio_method()
		ifconfigme__self_ip_str = extern_service_ifconfigme_method()
		print(f"DISCOVERED - self IPs:")
		print(f"dns         - {dns__self_ip_str}")
		print(f"ipinfo.io   - {ipinfo__self_ip_str}")
		print(f"ifconfig.me - {ifconfigme__self_ip_str}")
		
		# some users of this function expect the list of self-ips to be
		# unique and will cause errors otherwise.
		ips_no_dups_lst = list(set([
			dns__self_ip_str,
			ipinfo__self_ip_str,
			ifconfigme__self_ip_str
		]))

		return ips_no_dups_lst
	
	#---------------------------------------------------
	


	# CACHE
	cache_path_str = "self_ip_cache.txt"
	if os.path.isfile(cache_path_str):
		f=open("self_ip_cache.txt", "r")
		ips_str = f.read()
		f.close()



		ips_lst = []

		for ip_str in ips_str.split("\n"):

			# if text is a valid IP 
			if len(ip_str.split(".")) == 4:
				ips_lst.append(ip_str)

		print("CACHED - self IP:")
		print(ips_lst)

		return ips_lst

	# DISCOVER
	else:
		ips_lst = discover_ip()
			
		return ips_lst



	


# GloFlow application and media management/publishing platform
# Copyright (C) 2023 Ivan Trajkovic
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

import json
import boto3
import psycopg2

#---------------------------------------------------------------------------------
def init_db_client(p_db_name_str,
	p_env_str):
	
	db_user_str, db_pass_str, db_host_str, db_port_str = gf_postgresql_db_creds(p_env_str)

	db_client = psycopg2.connect(
		host     = db_host_str,
		port     = int(db_port_str),
		database = p_db_name_str,
		user     = db_user_str,
		password = db_pass_str)

	return db_client

#---------------------------------------------------------------------------------
def gf_postgresql_db_creds(p_db_env_str):

	secrets_client = boto3.client('secretsmanager',
		region_name="us-east-1")
	
	db_host_port_str = secrets_client.get_secret_value(SecretId=f"gf_rds_host_{p_db_env_str}")["SecretString"]
	db_creds_str     = secrets_client.get_secret_value(SecretId=f"gf_rds_creds_{p_db_env_str}")["SecretString"]
	print("db RDS creds fetched from aws secrets_manager...")

	db_creds_map = json.loads(db_creds_str)
	db_user_str = db_creds_map["username"]
	db_pass_str = db_creds_map["password"]

	db_host_str, db_port_str = db_host_port_str.split(":")

	return db_user_str, db_pass_str, db_host_str, db_port_str

#---------------------------------------------------------------------------------
def table_exists(p_table_name_str,
	p_db_cursor):
	p_db_cursor.execute(f"SELECT EXISTS(SELECT * FROM information_schema.tables WHERE table_name = '{p_table_name_str}')")
	exists_bool = p_db_cursor.fetchone()[0]
	return exists_bool


# GloFlow application and media management/publishing platform
# Copyright (C) 2020 Ivan Trajkovic
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

import os, sys
modd_str = os.path.abspath(os.path.dirname(__file__)) # module dir

import numpy as np
import matplotlib.pyplot as plt

import tensorflow as tf
import tensorflow.keras.layers
tf.compat.v1.enable_eager_execution() # EAGER_EXEC

import tensorflow.keras.utils

sys.path.append("%s/utils"%(modd_str))
import gf_ml_test_data

#----------------------------------------------
def main():

    print(tf.executing_eagerly())

    #----------------------------------------------
    def tf_set_cpu_target():
        cpu_devices_lst = tf.config.experimental.list_physical_devices(device_type="CPU")
        print(cpu_devices_lst)

        tf.config.experimental.set_visible_devices(devices=cpu_devices_lst, device_type="CPU")
        
        # enable logging of target devices used for execution of ops
        tf.debugging.set_log_device_placement(True)

    #----------------------------------------------

    tf_set_cpu_target()

    input_img_batch       = gf_ml_test_data.load_data()
    input_img_transformed = np.squeeze(input_img_batch, 0) # drop 1st (batch) dimension


    basic_conv_model(input_img_batch)



    test_conv_layer(input_img_batch)



#----------------------------------------------
def basic_conv_model(p_input_img_batch):
            
    input_l = tensorflow.keras.layers.Input(shape=(64, 64, 3))


    tf_conv_l = tensorflow.keras.layers.Conv2D(filters=8,
        kernel_size = (5, 5),
        padding     = "same",
        activation  = "relu")(input_l)

    maxpool_l = tensorflow.keras.layers.MaxPooling2D((2, 2))(tf_conv_l)




    # o = tf_conv_l(p_input_img_batch)
    model = tf.keras.models.Model(inputs=input_l, outputs=maxpool_l)
    model.summary()

    # MODEL_COMPILE
    model.compile(optimizer = "adam",
        loss    = tf.keras.losses.SparseCategoricalCrossentropy(from_logits=True),
        metrics = ["accuracy"])

    # MODEL_PREDICT
    y = model.predict(p_input_img_batch)
    print(y[0].shape)



    # CONV_LAYER_OUTPUT
    tf_conv_l_ouput_fn = tf.keras.backend.function([model.layers[0].input],
        [model.layers[1].output])




    conv_output = tf_conv_l_ouput_fn(p_input_img_batch)
    print(conv_output[0].shape)


    tf_conv_l_output = tf_conv_l_ouput_fn(p_input_img_batch)[0]





    # PLOT_MODEL
    tensorflow.keras.utils.plot_model(model, to_file="%s/data/output/tf_model.png"%(modd_str))


    #----------------------------------------------
    def plot_imgs():
        
        fig  = plt.figure(constrained_layout=False, figsize=(10, 5))
        spec = fig.add_gridspec(ncols=3, nrows=2)


        ax0 = fig.add_subplot(spec[:, 0]) # across all rows and first column
        ax1 = fig.add_subplot(spec[:, 1]) # across all rows and second column
        ax2 = fig.add_subplot(spec[0, 2]) # accross first row and third column


        # fig, (ax0, ax1, ax2) = fig.subplots(ncols=3, nrows=2, figsize=(10, 5))


        # IMPORTANT!! - transpose() - output of a convolutional layer and maxpool layer is transposed:
        #               (channels_num, filter_w, filter_h, filters_num)
        #               but to get individual filter values, we want:
        #               (filters_num, filter_w, filter_h, channels_num)

        first_filter = tf_conv_l_output.transpose()[0].squeeze()
        first_maxpool = y[0].transpose()[0]

        ax0.set_title("input image $x$", fontsize=8)
        ax0.imshow(p_input_img_batch[0].astype("uint8"))

        ax1.set_title("Conv2D layer output image", fontsize=8)
        ax1.imshow(first_filter.squeeze())

        ax2.set_title("MaxPooling2D layer output image", fontsize=8)
        ax2.imshow(first_maxpool.astype("uint8"))

        plt.show()

    #----------------------------------------------
    plot_imgs()

#----------------------------------------------
def test_conv_layer(p_input_img_batch):

    # CONVOLUTION_LAYER
    tf_conv_l = tensorflow.keras.layers.Conv2D(filters=3,
        kernel_size = (5, 5),
        padding     = "same",
        input_shape = (None, None, 3),
        activation  = "relu")

    img_in_np  = p_input_img_batch[0]
    img_out    = tf_conv_l(p_input_img_batch)
    img_out_np = img_out[0].numpy()

    # PLOT_FILTERS
    plot_conv_layer(tf_conv_l)

    #----------------------------------------------
    def plot_imgs():
        fig, (ax0, ax1) = plt.subplots(ncols=2, figsize=(10, 5))

        ax0.imshow(img_in_np.astype("uint8"))
        ax1.imshow(img_out_np.astype("uint8"))

        plt.show()
    
    #----------------------------------------------
    plot_imgs()


    
#----------------------------------------------
def plot_conv_layer(p_tf_conv_l):




    print("conv params #  - %s"%(p_tf_conv_l.count_params()))
    print("conv filters # - %s"%(len(p_tf_conv_l.get_weights())))

    print(dir(p_tf_conv_l))
    print(p_tf_conv_l.filters)
    print(p_tf_conv_l.rank)



    filters_num_int = p_tf_conv_l.filters

    vars_lst = p_tf_conv_l.get_weights()
    w        = vars_lst[0]
    b        = vars_lst[1]

    print(w.shape)
    print(b.shape)
    


    print(w.shape)


    # w             - (width, height, channels_num, filters_num)
    # w.transpose() - (filters_num, width, height, channels_num)
    # [3, 0, 1, 2]  - move the last axis (filters_num) to be the first
    n = np.transpose(w, axes=[3, 0, 1, 2])


    print(n.shape)
 
    
    print(n.shape)
    print(n[0].shape)

    print(n[0])

    print(p_tf_conv_l)




    # PLOT_FILTERS
    fig, axis_tpl = plt.subplots(ncols=filters_num_int, figsize=(10, 5))
    for i, ax in enumerate(axis_tpl):

        filter_weights = n[i]
        ax.set_title("filter %s"%(i))



        ax.imshow(filter_weights)



    plt.show()

#----------------------------------------------
main()


from gf_core import gf_core_id, gf_core_sql_db
from gf_apps.gf_images.gf_images_client import gf_images_client
from gf_observe import gf_extern_load
from gf_ml import gf_llm_core
from gf_extern_services.gf_aws import gf_aws_ec2, gf_aws_s3, gf_aws_secrets, gf_aws_route53

print("gloflow...")
version = "0.1.18"

#-------------------------
# EXPORT_API
# simplified unified GF Py API with flattened namespace

# DB
class db():
    init         = gf_core_sql_db.init_db_client
    table_exists = gf_core_sql_db.table_exists

# IMAGES
class images():
    add = gf_images_client.add_image

# CORE
class core():
    create_id = gf_core_id.create

# OBSERVE
class observe():
    init           = gf_extern_load.init
    ext_load       = gf_extern_load.observe
    get_cached     = gf_extern_load.get_cached
    get_cached_group = gf_extern_load.get_cached_group
    relate           = gf_extern_load.relate

# ML
class ml():
    run_llm = gf_llm_core.run_llm

# EXTERN SERVICES
class extern():
    aws_ec2 = gf_aws_ec2
    aws_s3  = gf_aws_s3
    aws_secrets = gf_aws_secrets
    aws_route53 = gf_aws_route53


# GloFlow application and media management/publishing platform
# Copyright (C) 2019 Ivan Trajkovic
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

import json
import gf_image

#---------------------------------------------------
#->:Bool
def db_image_exists(p_image_id_str,
	p_db_context_map,
	p_log_fun,
	p_db_type_str         = 'mongo',
	p_mongo_db_name_str   = 'prod_db',
	p_mongo_coll_name_str = 'images'):

	#-----------
	# ADD!! - use redis as a cache for mongo data
	if p_db_type_str == 'redis':
		key_str = 'img:%s'%(p_image_id_str)
		if p_db_context_map['redis_client'].exists(key_str):
			return True
		else:
			return False

	#-----------
	# MONGO
	elif p_db_type_str == 'mongo':
		mongo_client = p_db_context_map['mongodb_client']
		data_coll    = mongo_client[p_mongo_db_name_str][p_mongo_coll_name_str] #db/collection name
		
		#find().limit() - drastically faster for existence checking then findOne()
		#{"_id":1}      - only return the "_id" field
		if data_coll.find({"id_str": p_image_id_str},{"_id":1}).limit(1).count() == 0: return False
		else: return True
	
	#-----------

#---------------------------------------------------
#->:GFimage|None

def db_get(p_image_id_str,
	p_db_context_map,
	p_log_fun,
	p_db_type_str         = 'mongo',
	p_mongo_db_name_str   = 'prod_db',
	p_mongo_coll_name_str = 'images'):
	
	#---------------
	# DB_MONGO
	if p_db_type_str == 'mongo':
		
		# SCALING!! - image_exists() does a full query to mongo
		#             investigate further
		if db_image_exists(p_image_id_str,
			p_db_context_map,
			p_log_fun,
			p_db_type_str = 'mongo'):

			mongo_client       = p_db_context_map['mongodb_client']
			gf_images_coll     = mongo_client[p_mongo_db_name_str][p_mongo_coll_name_str]
			raw_image_info_map = gf_images_coll.find({"id_str": p_image_id_str})[0]
			image_info_map     = gf_image.deserialize(raw_image_info_map, p_log_fun)
		else:
			return None

	#---------------

	# create() - does verification and adt construction
	gf_image = gf_image.create(image_info_map, p_db_context_map, p_log_fun)
	assert isinstance(gf_image, gf_image.GFimage)
	
	return gf_image

#---------------------------------------------------
def db_put(p_image,
	p_db_context_map,
	p_log_fun,
	p_db_type_str         = 'mongo',
	p_mongo_db_name_str   = 'prod_db',
	p_mongo_coll_name_str = 'images'):
	assert isinstance(p_image, gf_image.GFimage)
	
	image_info_map = gf_image.serialize(p_image,
		p_log_fun)
	#---------------
	# DB
	if p_db_type_str == 'mongo':
		mongo_client = p_db_context_map['mongodb_client']
		data_coll    = mongo_client[p_mongo_db_name_str][p_mongo_coll_name_str]

		# spec          - a dict specifying elements which must be present for a document to be updated
		# upsert = True - insert doc if it doesnt exist, else just update
		data_coll.update({'id_str':p_image.id_str}, # spec
			image_info_map, 
			upsert = True)
		
	#---------------
	
#---------------------------------------------------	
#->:List<:Image_ADT>
def db_get_all(p_db_context_map,
	p_log_fun,
	p_db_type_str         = 'mongo',
	p_mongo_db_name_str   = 'prod_db',
	p_mongo_coll_name_str = 'images'):

	#----------
	# MONGO
	if p_db_type_str == 'mongo':
		True
		
	#----------


# GloFlow application and media management/publishing platform
# Copyright (C) 2019 Ivan Trajkovic
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

import os
import hashlib
import json
import urlparse

import gf_image_verify

#---------------------------------------------------
class GFimage():
	def __init__(self, p_props_map):
		
		self.id_str    = p_props_map['id_str']
		self.title_str = p_props_map['title_str']
		
		#---------------
		# when the image comes from an external url (as oppose to it being 
		# created internally, or uploaded directly to the system)
		self.origin_url_str = p_props_map.get('origin_url_str', None)

		# actual path on the filesystem, of the fullsized image gotten from origin_url_str
		self.original_file_internal_uri_str = p_props_map.get('original_file_internal_uri_str', None)
		
		#---------------
		# relative url's
		# '/images/image_name.*'
		
		self.thumbnail_small_url_str  = p_props_map.get('thumbnail_small_url_str',  None)
		self.thumbnail_medium_url_str = p_props_map.get('thumbnail_medium_url_str', None)
		self.thumbnail_large_url_str  = p_props_map.get('thumbnail_large_url_str',  None)
		
		#---------------
		self.format_str = p_props_map['format_str'] #"jpeg"|"png"|"gif"
		self.width_str  = p_props_map.get('width_str',  None)
		self.height_str = p_props_map.get('height_str', None)
		
		#---------------
		self.dominant_color_hex_str = p_props_map.get('dominant_color_hex_str', None)
		
		self.tags_lst = p_props_map.get('tags_lst', None)
		
#---------------------------------------------------
#->:GFimage

def create(p_image_info_map,
	p_db_context_map,
	p_log_fun):
	
	new_image_info_map = gf_image_verify.verify_image_info(p_image_info_map,
		p_db_context_map,
		p_log_fun)

	if not p_image_info_map.has_key('id_str'):
		image_id_str = create_id(new_image_info_map['title_str'],
			new_image_info_map['format_str'],
			p_log_fun)
		new_image_info_map['id_str'] = image_id_str
		
	gf_image = GFimage(new_image_info_map)
	return gf_image

#---------------------------------------------------
def create_id_from_url(p_image_url_str, p_log_fun):
	
	# urlparse() - used so that any possible url query parameters are not used in the 
	#              os.path.basename() result
	image_path_str           = urlparse.urlparse(p_image_url_str).path
	image_file_name_str      = os.path.basename(image_path_str)
	image_title_str, ext_str = os.path.splitext(image_file_name_str)
	normalized_ext_str       = gf_image_verify.check_image_format_str(ext_str.lower().strip('.'), p_log_fun)
	
	image_id_str = create_id(image_path_str,
		normalized_ext_str, # p_image_format_str,
		p_log_fun)
	return image_id_str

#---------------------------------------------------
# p_image_type_str - :String - 'jpeg'|'gif'|'png'

#->:String(image_id_str)
def create_id(p_image_path_str,
	p_image_format_str,
	p_log_fun):
	assert isinstance(p_image_path_str, str)
	
	assert p_image_format_str == 'jpeg' or \
		p_image_format_str == 'jpg'  or \
		p_image_format_str == 'gif'  or \
		p_image_format_str == 'png'

	m = hashlib.md5()
	m.update(p_image_path_str)
	m.update(p_image_format_str)
	
	#-------------------
	# hexdigest() - Like digest() except the digest is returned as a string of 
	#               double length, containing only hexadecimal digits. This may be used to 
	#               exchange the value safely in email or other non-binary environments.
	image_id_str = m.hexdigest()

	#-------------------
	
	return image_id_str

#---------------------------------------------------
#->:Map(image_info_map)
def serialize(p_gf_image, p_log_fun):

	image_info_map = {
		'id_str':    p_gf_image.id_str,
		'title_str': p_gf_image.title_str,
		
		'origin_url_str':                 p_gf_image.origin_url_str,
		'original_file_internal_uri_str': p_gf_image.original_file_internal_uri_str,
		
		'thumbnail_small_url_str':  p_gf_image.thumbnail_small_url_str,
		'thumbnail_medium_url_str': p_gf_image.thumbnail_medium_url_str,
		'thumbnail_large_url_str':  p_gf_image.thumbnail_large_url_str,
		
		'format_str': p_gf_image.format_str,
		'width_str':  p_gf_image.width_str,
		'height_str': p_gf_image.height_str,
		
		'dominant_color_hex_str': p_gf_image.dominant_color_hex_str,
		'tags_lst':               p_gf_image.tags_lst
	}
	
	return image_info_map

#---------------------------------------------------
#->:Map(image_info_map)

def deserialize(p_raw_image_info_map, p_log_fun):
	assert isinstance(p_raw_image_info_map, dict)
	
	image_info_map = p_raw_image_info_map
	return image_info_map


# GloFlow application and media management/publishing platform
# Copyright (C) 2020 Ivan Trajkovic
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

import os, sys
modd_str = os.path.abspath(os.path.dirname(__file__)) # module dir

import time
from colored import fg, bg, attr

sys.path.append("%s/../../../go/gf_core/py"%(modd_str))
import gf_core_cli

#---------------------------------------------------
# ROOT
def generate__ca_root(p_output_files_base_str,
	p_config__file_path_str,
	p_sudo_bool = False):
	assert os.path.isfile(p_config__file_path_str)
	assert p_config__file_path_str.endswith(".json")
	
	print("%sGENERATE ROOT_CA%s"%(fg("yellow"), attr(0)))

	c_lst = []

	#-----------------
	# GENERATE
	if p_sudo_bool:
		c_lst.append("sudo")
	
	c_lst.extend([
		"cfssl gencert",
		"-initca", # "-initca" - initialise new CA
		p_config__file_path_str
	])

	#-----------------
	c_lst.append("|")

	#-----------------
	# SAVE_TO_FS
	if p_sudo_bool:
		c_lst.append("sudo")

	c_lst.extend([
		# "-bare" - the response from CFSSL is not wrapped in the API standard response
		"cfssljson -bare %s"%(p_output_files_base_str)
	])

	#-----------------
	c_str = " ".join(c_lst)
	_, _, return_code = gf_core_cli.run_cmd(c_str, p_env_map = None)
	if not return_code == 0:
		print("CLI failed...")
		exit()

	print("%sdone...%s"%(fg("green"), attr(0)))

#-------------------------------------------------------------
# INTERMEDIATE
def generate__ca_intermediate(p_output_files_base_str,
	p_root_ca_base_str,
	p_config__file_path_str,
	p_config_csr__file_path_str,
	p_profile_name_str = None,
	p_sudo_bool        = False):
	assert os.path.isfile(p_config__file_path_str)
	assert os.path.isfile(p_config_csr__file_path_str)

	print("%sGENERATE INTERMEDIATE_CA%s"%(fg("yellow"), attr(0)))

	# ROOT_CA
	root_ca_cert__file_path_str = "%s.pem"%(p_root_ca_base_str)
	root_ca_key__file_path_str  = "%s-key.pem"%(p_root_ca_base_str)

	c_lst = []
	
	#-----------------
	# GENERATE
	if p_sudo_bool:
		c_lst.append("sudo")

	c_lst.extend([
		"cfssl gencert",

		# ROOT_CA
		"-ca %s"%(root_ca_cert__file_path_str),
		"-ca-key %s"%(root_ca_key__file_path_str),

		"-config %s"%(p_config__file_path_str)
	])

	if not p_profile_name_str == None:
		c_lst.append("-profile %s"%(p_profile_name_str))

	c_lst.append(p_config_csr__file_path_str)

	#-----------------
	c_lst.append("|")

	#-----------------
	# SAVE_TO_FS
	if p_sudo_bool:
		c_lst.append("sudo")

	c_lst.extend([
		"cfssljson -bare %s"%(p_output_files_base_str)
	])
	
	#-----------------
	c_str = " ".join(c_lst)
	_, _, return_code = gf_core_cli.run_cmd(c_str, p_env_map = None)
	if not return_code == 0:
		print("CLI failed...")
		exit()

	print("%sdone...%s"%(fg("green"), attr(0)))

#-------------------------------------------------------------
# LEAF
def generate__cert_leaf(p_output_files_base_str,
	p_ca_base_str,
	p_config__file_path_str,
	p_hostname_str = None,
	p_sudo_bool    = False):

	# INTERMEDIATE_CA
	ca_cert__file_path_str = "%s.pem"%(p_ca_base_str)
	ca_key__file_path_str  = "%s-key.pem"%(p_ca_base_str)

	c_lst = []
	
	#-----------------
	# GENERATE

	c_lst.append('''echo "{}"''')
	c_lst.append("|")

	if p_sudo_bool:
		c_lst.append("sudo")

	c_lst.extend([
		"cfssl gencert",

		# INTERMEDIATE_CA
		"-ca=%s"%(ca_cert__file_path_str),
		"-ca-key=%s"%(ca_key__file_path_str),
		"-config=%s"%(p_config__file_path_str),
	])

	if not p_hostname_str == None:
		c_lst.append('''-hostname="%s"'''%(p_hostname_str))

	# pass-in CSRJSON from stdin from the initial 'echo "{}"'
	c_lst.append("-")

	#-----------------
	c_lst.append("|")

	#-----------------
	# SAVE_TO_FS
	if p_sudo_bool:
		c_lst.append("sudo")

	c_lst.append("cfssljson -bare %s"%(p_output_files_base_str))

	#-----------------
	c_str = " ".join(c_lst)
	_, _, return_code = gf_core_cli.run_cmd(c_str, p_env_map = None)
	if not return_code == 0:
		print("CLI failed...")
		exit()

	print("%sdone...%s"%(fg("green"), attr(0)))

#-------------------------------------------------------------
def archive_if_exists(p_files_base_str, p_sudo_bool = False):
	
	dir_str       = os.path.abspath(os.path.dirname(p_files_base_str))
	file_base_str = os.path.basename(p_files_base_str)

	if p_sudo_bool: sudo_str = "sudo"
	else:           sudo_str = ""

	# list all files in target dir
	# "-1"  - force output to be one entry per line
	# "^%s\.\|^%s-key\." - pattern matches the file_base only at the start of the line with a postfix "." or "-key."
	stdout_str, _, return_code = gf_core_cli.run_cmd("%s ls -1 %s | grep '^%s\.\|^%s-key\.'"%(sudo_str, dir_str, file_base_str, file_base_str),
		p_env_map = None,
		p_print_output_bool = True)
	
	if stdout_str == "":
		return True

	stdout_clean_str = stdout_str.strip()
	lines_lst        = stdout_clean_str.split("\n")

	if len(lines_lst) > 0:

		# IMPORTANT!! - ask use if they want to recreate/archive existing certs. if they dont
		#               dont archive and return False
		print("CERT ALREADY EXISTS - %s"%(file_base_str))
		if not gf_core_cli.confirm("recreate cert (and archive old)?"):
			return False

		archive_time = time.time()

		# process each file that needs to be archivedp_ca_intermediate__output_files_base_str
		for l in lines_lst:
			file_name_str = l.split()[-1:][0]
			file_path_str = "%s/%s"%(dir_str, file_name_str)

			# ARCHIVE_FILE - rename the file
			c = "%s mv %s %s/old_%s__%s"%(sudo_str, file_path_str, dir_str, archive_time, file_name_str)
			_, _, return_code = gf_core_cli.run_cmd(c, p_env_map = None)
			if not return_code == 0:
				print("CLI failed...")
				exit()
	
	return True

#-------------------------------------------------------------
def combine_ca_chain(p_output_files_base_str,
	p_cert_chain__file_paths_lst,
	p_sudo_bool = False):

	output_file_str = "%s_full.pem"%(p_output_files_base_str)

	c_lst = []
	if p_sudo_bool:
		c_lst.append("sudo")

	c_lst.extend([
		"touch %s"%(output_file_str),
		"&&"
	])
	
	if p_sudo_bool:
		c_lst.append("sudo")


	for f in p_cert_chain__file_paths_lst:
		assert f.endswith(".pem")

	# child-certs go first in the list, before their parent certs.
	c_lst.extend(["bash -c 'cat %s > %s'"%(
		" ".join(p_cert_chain__file_paths_lst),
		output_file_str
	)])

	_, _, return_code = gf_core_cli.run_cmd(" ".join(c_lst), p_env_map = None)
	if not return_code == 0:
		print("CLI failed...")
		exit()

	return output_file_str


# GloFlow application and media management/publishing platform
# Copyright (C) 2024 Ivan Trajkovic
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

import json
from icecream import ic
import pandas as pd
import gloflow as gf

table_name__extern_load_str = "gf_extern_load"

#---------------------------------------------------------------------------------
def get_load_type_counts(p_db_client):
	query_str = f'''
		SELECT load_type, COUNT(*)
		FROM {table_name__extern_load_str}
		GROUP BY load_type
		ORDER BY load_type;
	'''
	df = pd.read_sql(query_str, p_db_client)
	return df

#---------------------------------------------------------------------------------
def relate_observations(p_target_observation_id_int,
	p_related_observations_ids_lst,
	p_db_client):
	assert(isinstance(p_target_observation_id_int, int))
	assert(isinstance(p_related_observations_ids_lst, list))
	for o in p_related_observations_ids_lst:
		assert(isinstance(o, int))

	cur = p_db_client.cursor()

	query_str = f'''
		UPDATE {table_name__extern_load_str}
		SET related_observations = %s
		WHERE id = %s
	'''

	cur.execute(query_str, 
		(
			p_related_observations_ids_lst,
			p_target_observation_id_int
		))
	
	p_db_client.commit()
	cur.close()

#---------------------------------------------------------------------------------
# GET_GROUP

def get_group(p_load_type_str,
	p_part_key_str,
	p_source_domain_str,
	p_group_id_str,
	p_db_client):
	
	cur = p_db_client.cursor()
	query_str = f'''
		SELECT
			id,
			url,
			resp_cache_file_path,
			meta_map
		FROM {table_name__extern_load_str}
		WHERE
			load_type     = %s AND
			part_key      = %s AND
			source_domain = %s AND
			group_id      = %s

		ORDER BY fetch_datetime DESC;
	'''

	params_lst = [
		p_load_type_str,
		p_part_key_str,
		p_source_domain_str,
		p_group_id_str
	]

	cur.execute(query_str, params_lst)
	results_lst = cur.fetchall()

	if results_lst is None:
		return None
	else:
		ic(results_lst)

		loads_lst = []
		for row in results_lst:
			load_map = {
				"id":                   row[0],
				"url":                  row[1],
				"resp_cache_file_path": row[2],
				"meta_map":             row[3]
			}
			loads_lst.append(load_map)

		return loads_lst

#---------------------------------------------------------------------------------
# GET_LATEST_GROUP_ID

def get_latest_group_id(p_load_type_str,
	p_source_domain_str,
	p_db_client):

	cur = p_db_client.cursor()
	query_str = f'''
		SELECT group_id
		FROM {table_name__extern_load_str}
		WHERE
			load_type     = %s AND
			source_domain = %s AND
			group_id IS NOT NULL

		ORDER BY fetch_datetime DESC
	'''

	cur.execute(query_str, [
		p_load_type_str,
		p_source_domain_str
	])
	result_tpl = cur.fetchone()
	group_id_str = result_tpl[0]

	return group_id_str

#---------------------------------------------------------------------------------
# GET_LATEST_LOAD

def get_latest(p_load_type_str,
	p_part_key_str,
	p_source_domain_str,
	p_db_client,
	p_group_id_str=None):

	where_sql_str = "load_type = %s AND part_key = %s AND source_domain = %s"
	params_lst = [
		p_load_type_str,
		p_part_key_str,
		p_source_domain_str
	]
	if p_group_id_str is not None:
		where_sql_str += " AND group_id = %s"
		params_lst.append(p_group_id_str)

	cur = p_db_client.cursor()
	query_str = f'''
		SELECT
			id,
			url,
			resp_cache_file_path,
			meta_map
		FROM {table_name__extern_load_str}
		WHERE {where_sql_str}
		ORDER BY fetch_datetime DESC
		LIMIT 1;
	'''

	cur.execute(query_str, params_lst)
	result_tpl = cur.fetchone()

	if result_tpl is None:
		return None
	else:
		ic(result_tpl)
		load_map = {
			"id":                   result_tpl[0],
			"url":                  result_tpl[1],
			"resp_cache_file_path": result_tpl[2],
			"meta_map":             result_tpl[3]
		}
		return load_map

#---------------------------------------------------------------------------------
# INSERT
def insert(p_load_type_str,
	p_part_key_str,
	p_source_domain_str,
	p_cache_s3_key_str,
	p_db_client,
	p_meta_map     = {},
	p_url_str      = None,
	p_group_id_str = None):
	assert(isinstance(p_meta_map, dict))

	cur = p_db_client.cursor()

	query_str = f'''INSERT INTO {table_name__extern_load_str} (
			load_type,
			part_key,
			url,
			resp_cache_file_path,
			meta_map,
			source_domain,
			group_id
		)
		VALUES (%s, %s, %s, %s, %s, %s, %s)
		RETURNING id
	'''

	cur.execute(query_str, 
		(
			p_load_type_str,
			p_part_key_str,
			p_url_str,
			p_cache_s3_key_str,
			json.dumps(p_meta_map),
			p_source_domain_str,
			p_group_id_str
		))

	id_int = cur.fetchone()[0]
	p_db_client.commit()
	cur.close()

	return id_int

#---------------------------------------------------------------------------------
# INIT
def init(p_db_client):

	cur = p_db_client.cursor()
	
	if not gf.db.table_exists(table_name__extern_load_str, cur):
		
		# source_domain  - domain on which this ad was discovered
		# fetch_datetime - time when the GF system stored this item

		sql_str = f"""
			CREATE TABLE {table_name__extern_load_str} (
			
				id SERIAL PRIMARY KEY,
				
				-- what type of extern loading is done
				-- model, model_variant, etc.
				load_type VARCHAR(255),

				-- -----------------------
				-- PARTITION_KEY
				-- string representing partition key
				part_key VARCHAR(255),

				-- -----------------------
				-- URL
				-- for html data this is the URL of the page, for json
				-- returns it might be the URL of the API endpoint.
				-- for other types of data it might be None.

				url VARCHAR(1000),
				
				-- -----------------------
				-- RELATED_OBSERVATIONS
				-- these are observations that this observations depends on or is related to.
				-- there can be multiple observations that can be used by this observation.

				related_observations INT[],

				-- -----------------------
				-- RESPONSE
				resp_cache_file_path TEXT,

				-- -----------------------
				-- META
				-- various metadata that can be attached to a load event.
				meta_map JSONB,
				
				-- -----------------------
				-- GROUP_ID
				-- allows for grouping of observations.
				group_id VARCHAR(255),

				-- -----------------------

				source_domain  VARCHAR(255),
				fetch_datetime TIMESTAMP DEFAULT NOW()
			);
		"""
		cur.execute(sql_str)
		p_db_client.commit()


# GloFlow application and media management/publishing platform
# Copyright (C) 2019 Ivan Trajkovic
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

import os, sys
modd_str = os.path.abspath(os.path.dirname(__file__))

import os
from colored import fg, bg, attr
from bs4 import BeautifulSoup

sys.path.append("%s/../../gf_core"%(modd_str))
import gf_core_cli

#--------------------------------------------------
# RUN_IN_CONTAINER
def run_in_cont(p_app_str,
	p_page_name_str=None):

	repo_local_path_str = os.path.abspath(f'{modd_str}/../../../../gloflow').strip()

	py_cmd_lst = [
		"python3", "-u", "/home/gf/ops/cli__build.py", "-run=build_web", "-build_outof_cont",
		f"-app={p_app_str}"
	]
	if not p_page_name_str == None:
		py_cmd_lst.append(f"-page_name={p_page_name_str}")

	cmd_lst = [
		"sudo", "docker", "run",
		"--rm", # remove after exit 
		"-v", f"{repo_local_path_str}:/home/gf", # mount repo into the container
		"glofloworg/gf_builder_web:latest",
	]
	cmd_lst.extend(py_cmd_lst)
	
	p = gf_core_cli.run__view_realtime(cmd_lst, {},
		"gf_build_web", "green")

	p.wait()

#---------------------------------------------------
def build(p_apps_names_lst,
	p_apps_meta_map,
	p_log_fun,
	p_page_name_str=None):
	assert isinstance(p_apps_names_lst, list)
	assert len(p_apps_names_lst) > 0
	assert isinstance(p_apps_meta_map, dict)

	#---------------------------------------------------
	def individual_page(p_page_name_str, p_page_info_map):
		print(f"page name {fg('yellow')}{p_page_name_str}{attr(0)}")

		build_dir_str = os.path.abspath(p_page_info_map["build_dir_str"])

		build_copy_dir_str     = p_page_info_map.get("build_copy_dir_str", None)
		build_copy_bas_dir_str = None
		if not build_copy_dir_str == None:
			build_copy_bas_dir_str = os.path.abspath(build_copy_dir_str)

		build_page(p_page_name_str,
			build_dir_str,
			build_copy_bas_dir_str,
			p_page_info_map,
			p_log_fun)

	#---------------------------------------------------
	for app_str in p_apps_names_lst:
		
		#-----------------
		# META
		if not app_str in p_apps_meta_map.keys():
			p_log_fun("ERROR", f"supplied app ({app_str}) does not exist in gf_web_meta")
			return

		app_map = p_apps_meta_map[app_str]

		#-----------------
		# BUILD PAGES - build each page of the app.
		#               no specific page was picked for build
		if p_page_name_str == None:
			
			for page_name_str, page_info_map in app_map["pages_map"].items():
				individual_page(page_name_str, page_info_map)
				
		else:
			page_info_map = app_map["pages_map"][p_page_name_str]

			individual_page(p_page_name_str, page_info_map)

#---------------------------------------------------
# BUILD_PAGE

def build_page(p_page_name_str,
	p_build_dir_str,
	p_build_copy_dir_str,
	p_page_info_map,
	p_log_fun):
	
	print("")
	p_log_fun("INFO", "%s>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>%s"%(fg("orange_red_1"), attr(0)))
	p_log_fun("INFO", "             %sBUILD PAGE%s - %s%s%s"%(fg("cyan"), attr(0), fg("orange_red_1"), p_page_name_str, attr(0)))
	p_log_fun("INFO", "%s>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>%s"%(fg("orange_red_1"), attr(0)))
	print("")

	p_log_fun("INFO", f"build_dir_str - {p_build_dir_str}")
	assert isinstance(p_build_dir_str, str)

	# make build dir if it doesnt exist
	if not os.path.isdir(p_build_dir_str):
		gf_core_cli.run(f"mkdir -p {p_build_dir_str}")

	if "main_html_path_str" in p_page_info_map.keys():

		main_html_path_str = os.path.abspath(p_page_info_map["main_html_path_str"])
		
		print(f"main html path - {main_html_path_str}")

		assert os.path.isfile(main_html_path_str)
		assert main_html_path_str.endswith(".html")
		assert ".".join(os.path.basename(main_html_path_str).split(".")[:-1]) == p_page_name_str
		p_log_fun("INFO", f"main_html_path_str - {main_html_path_str}")

	# URL_BASE
	url_base_str = ""
	if "url_base_str" in p_page_info_map.keys():
		url_base_str = p_page_info_map["url_base_str"]
		p_log_fun("INFO", f"url_base_str - {url_base_str}")

	if "main_html_path_str" in p_page_info_map.keys():	
		f = open(main_html_path_str, "r")
		main_html_str = f.read()
		f.close()

		soup = BeautifulSoup(main_html_str)
		
	#---------------------------------------------------
	# JS
	
	def process_scripts():
		scripts_dom_nodes_lst = soup.findAll("script")

		# if there are scripts detected in the page
		if len(scripts_dom_nodes_lst) > 0:
			js_libs_build_dir_str = f"{p_build_dir_str}/js/lib"
			gf_core_cli.run(f"mkdir -p {js_libs_build_dir_str}") # create dir and all parent dirs

		main_html_dir_path_str = os.path.dirname(main_html_path_str)
		assert os.path.isdir(main_html_dir_path_str)

		for script_dom_node in scripts_dom_nodes_lst:

			# some <script> tags might just contain source code, and not reference an external JS file.
			# using .get("src") instead of ["src"] because the "src" DOM attribute might not be present.
			if script_dom_node.get("src") == None:
				continue

			src_str = script_dom_node["src"]

			if src_str.startswith("http://") or src_str.startswith("https://"):
				print("EXTERNAL_URL - DO NOTHING")
				continue
			

			
			
			local_path_str = os.path.abspath(f"{main_html_dir_path_str}/{src_str}")
			print(local_path_str)
			assert os.path.isfile(local_path_str)

			#-----------------
			if local_path_str.endswith(".ts"):
				p_log_fun("INFO", "%s------------ TYPESCRIPT --------------------------------%s"%(fg("yellow"), attr(0)))
				
				#---------------------------------------------------
				def build_typescript(p_out_file_str):
					
					cmd_lst = [
						
						"tsc",
						"--module system", # needed with the "--out" option
						"--target ES2020", # "--target es2017", # "--target es6",

						# Enables emit interoperability between CommonJS and ES Modules via creation of namespace objects for all imports
						# '--esModuleInterop',
						f"--outFile {p_out_file_str}",

						#---------------------
						# TSCONFIG 
						# f"-p {modd_str}/../../../web/tsconfig.json",
						
						#---------------------

						main_ts_file_str
					]
					cmd_str = " ".join(cmd_lst)
					print(cmd_str)
					
					_, _, return_code_int = gf_core_cli.run(cmd_str)

					if return_code_int > 0:
						print("ERROR!! - TypeScript Compilation failed!")
						exit(-1)

					# minify into the same file name as the Typescript compiler output
					target_dir_str = os.path.dirname(p_out_file_str)
					minify_js(p_out_file_str, [p_out_file_str], p_log_fun)

				#---------------------------------------------------

				main_ts_file_str       = local_path_str
				minified_file_name_str = "%s.min.js"%(".".join(os.path.basename(main_ts_file_str).split(".")[:-1]))
				minified_file_path_str = f"{p_build_dir_str}/js/{minified_file_name_str}"

				build_typescript(minified_file_path_str)

				# HTML_MODIFY - change the src in the html tag to the minified name, and url_base (dont leave relative path)
				script_dom_node["src"] = f"{url_base_str}/js/{minified_file_name_str}"

			#-----------------
			elif local_path_str.endswith(".js"):
				
				p_log_fun("INFO", "%s------------ JAVASCRIPT --------------------------------%s"%(fg("yellow"), attr(0)))

				# IMPORTANT!! - just copy the JS file to the final build dir
				gf_core_cli.run(f"cp {local_path_str} {js_libs_build_dir_str}")

				# HTML_MODIFY - change the src in the html tag, to include the url_base (dont leave relative path)
				script_dom_node["src"] = f"{url_base_str}/js/lib/{os.path.basename(local_path_str)}"

			#-----------------

	#---------------------------------------------------
	# CSS
	
	def process_css():

		p_log_fun("INFO", "%s------------ CSS ---------------------------------------%s"%(fg("yellow"), attr(0)))
		css_links_lst = soup.findAll("link", {"type": "text/css"})
		
		target_dir_str = f'{p_build_dir_str}/css/{p_page_name_str}'
		gf_core_cli.run(f'mkdir -p {target_dir_str}') #create dir and all parent dirs

		for css in css_links_lst:
			src_str = css["href"]
			
			assert src_str.endswith('.css') or src_str.endswith('.scss')

			if src_str.startswith('http://') or src_str.startswith('https://'):
				print('EXTERNAL_URL - DO NOTHING')
				continue

			# full paths are relative to the dir holding the main html file (app entry point)
			full_path_str = os.path.abspath(f'{os.path.dirname(main_html_path_str)}/{src_str}')
			print(full_path_str)
			assert os.path.isfile(full_path_str)
			
			# SASS
			if src_str.endswith('.scss'):
				css_file_name_str = os.path.basename(src_str).replace('.scss', '.css')
				final_src_str     = f'{target_dir_str}/{css_file_name_str}'
				gf_core_cli.run(f'sass {full_path_str} {final_src_str}')

				# HTML_MODIFY - change the src in the html tag, to include the url_base
				#               (dont leave relative path)
				css["href"] = f'{url_base_str}/css/{p_page_name_str}/{css_file_name_str}'

			# CSS
			else:
				
				gf_core_cli.run(f'cp {full_path_str} {target_dir_str}')

				# HTML_MODIFY - change the src in the html tag, to include the url_base (dont leave relative path)
				css["href"] = f'{url_base_str}/css/{p_page_name_str}/{os.path.basename(full_path_str)}'

	#---------------------------------------------------

	if "main_html_path_str" in p_page_info_map.keys():

		process_scripts()

		#-----------------
		# CSS
		css_process_bool = p_page_info_map.get("css_process_bool", True)
		if css_process_bool:
			process_css()

		#-----------------
		# CREATE_FINAL_MODIFIED_HTML - create the html template file in the build dir that contains all 
		#                              the modified urls for JS/CSS
		target_html_file_path_str = f'{p_build_dir_str}/templates/{p_page_name_str}/{p_page_name_str}.html'
		gf_core_cli.run(f'mkdir -p {os.path.dirname(target_html_file_path_str)}')

		f = open(target_html_file_path_str, 'w+')
		f.write(soup.prettify())
		f.close()

		#-----------------

	#-----------------
	# SUBTEMPLATES
	if "subtemplates_lst" in p_page_info_map.keys():
		process_subtemplates(p_page_name_str,
			p_build_dir_str,
			p_page_info_map,
			p_log_fun)

	#-----------------
	# IMPORTANT!! - do after build_copy_dir is created
	if "files_to_copy_lst" in p_page_info_map.keys():
		process_files_to_copy(p_page_info_map, p_log_fun)

	#-----------------
	# BUILD_COPY - this propety allows for the build dir of a page to be copied to some other dir after the build is complete.
	#              this has to run after all other build steps complete, so that it includes all the build artifacts.
	#
	# IMPORTANT!! - only some pages in some apps define this. gf_solo is one of these apps, it adds this property
	#               to the page defs of all other apps (since gf_solo includes all apps).
	if not p_build_copy_dir_str == None:
		print(f"copying {fg('green')}build{attr(0)} dir ({p_build_dir_str}) to {fg('yellow')}{p_build_copy_dir_str}{attr(0)}")
		
		gf_core_cli.run(f'mkdir -p {p_build_copy_dir_str}')
		gf_core_cli.run(f'cp -r {p_build_dir_str} {p_build_copy_dir_str}')

	#-----------------
	print("")
	p_log_fun("INFO", "%s>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>%s END"%(fg("orange_red_1"), attr(0)))
	print("")

#---------------------------------------------------
def process_files_to_copy(p_page_info_map, p_log_fun):
	assert isinstance(p_page_info_map, dict)

	print("")
	p_log_fun("INFO", "%s------------ COPY_FILES --------------------------------%s"%(fg("yellow"), attr(0)))
	print("")

	files_to_copy_lst = p_page_info_map["files_to_copy_lst"]
	assert isinstance(files_to_copy_lst, list)


	# COPY_FILES
	for file_to_copy_tpl in files_to_copy_lst:
		assert isinstance(file_to_copy_tpl, tuple)
		src_file_str, target_dir_str = file_to_copy_tpl

		
		
		# TARGET_DIR
		if not os.path.isdir(target_dir_str):
			gf_core_cli.run(f'mkdir -p {target_dir_str}')


		c_lst = ["cp"]


		if os.path.isdir(src_file_str):
			c_lst.append("-r")
		else:
			assert os.path.isfile(src_file_str)

		c_lst.append(src_file_str)
		c_lst.append(target_dir_str)

		c_str = " ".join(c_lst)
		gf_core_cli.run(c_str)

		if os.path.isfile(src_file_str):
			final_path_str = f"{target_dir_str}/{os.path.basename(src_file_str)}"
			assert os.path.isfile(final_path_str)

#---------------------------------------------------
def process_subtemplates(p_page_name_str,
	p_build_dir_str,
	p_page_info_map,
	p_log_fun):
	assert isinstance(p_page_name_str, str)
	assert os.path.isdir(p_build_dir_str)
	assert "subtemplates_lst" in p_page_info_map.keys()

	print("")
	p_log_fun("INFO", "%s------------ SUBTEMPLATES --------------------------------%s"%(fg("yellow"), attr(0)))
	print("")

	subtemplates_lst = p_page_info_map["subtemplates_lst"]
	assert isinstance(subtemplates_lst, list)

	# SUBTEMPLATES__BUILD_DIR
	target_subtemplates_build_dir_str = f"{p_build_dir_str}/templates/{p_page_name_str}/subtemplates"
	if not os.path.isdir(target_subtemplates_build_dir_str):
		gf_core_cli.run(f"mkdir -p {target_subtemplates_build_dir_str}")

	for s_path_str in subtemplates_lst:
		print(s_path_str)
		assert isinstance(s_path_str, str)
		assert os.path.isfile(s_path_str)
		assert s_path_str.endswith(".html")

		# SUBTEMPLATE__COPY
		gf_core_cli.run(f"cp {s_path_str} {target_subtemplates_build_dir_str}")

#---------------------------------------------------
def minify_js(p_js_target_file_str,
	p_js_files_lst,
	p_log_fun):
	p_log_fun("FUN_ENTER", "gf_web__build.minify_js()")

	cmd_lst = [
		"uglifyjs",
		f"--output {p_js_target_file_str}",
		" ".join(p_js_files_lst),
	]
	gf_core_cli.run(" ".join(cmd_lst))

#---------------------------------------------------


# GloFlow application and media management/publishing platform
# Copyright (C) 2019 Ivan Trajkovic
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

import sys, os
import argparse
import subprocess
import threading

from colored import fg, bg, attr
import delegator

#---------------------------------------------------
def run__view_realtime(p_cmd_lst,
	p_env_map,
	p_view__type_str,
	p_view__color_str):

	print(" ".join(p_cmd_lst))
	
	# When shell=True the shell is the child process, and the commands are its children.
	# So any SIGTERM or SIGKILL will kill the shell but not its child processes.
	# The best way I can think of is to use shell=False, otherwise when you kill
	# the parent shell process, it will leave a defunct shell process.
	# CMD also has to be a list here, since its not being passed in as a string
	# to the child shell.
	p = subprocess.Popen(p_cmd_lst, shell=False, stdout=subprocess.PIPE, bufsize=1,
		env=p_env_map)

	t = threading.Thread(target=read_process_std_stream, args=(p.stdout, p_view__type_str, p_view__color_str))
	t.start()

	return p

#---------------------------------------------------
# RUN
def run(p_cmd_str,
	p_env_map = {},
	p_exit_on_fail_bool = True,
	p_print_cmd_str     = True):

	# env map has to contains all the parents ENV vars as well
	p_env_map.update(os.environ)

	if p_print_cmd_str:
		print(f"{fg('yellow')}cmd{attr(0)} >>> {fg('green')}{p_cmd_str}{attr(0)}")

	p = subprocess.Popen(p_cmd_str,
		env     = p_env_map,
		shell   = True,
		stdout  = subprocess.PIPE,
		stderr  = subprocess.PIPE,
		bufsize = 1)

	t_o = threading.Thread(target=read_process_std_stream, args=(p.stdout, "stdout", "green"))
	t_o.start()

	t_e = threading.Thread(target=read_process_std_stream, args=(p.stderr, "stderr", "yellow"))
	t_e.start()

	p.wait()
	
	# wait for stdout/stderr printing threads to complete as well before returning from this function
	t_o.join()
	t_e.join()

	if p_exit_on_fail_bool:
		if not p.returncode == 0:

			print(f"ERROR!! - shell CMD ({p_cmd_str}) failed!")
			exit(p.returncode)

	return "", "", p.returncode

#-------------------------------------------------------------
def read_process_std_stream(p_std_stream,
	p_view_type_str,
	p_view_color_str):

	for line in iter(p_std_stream.readline, b''):
		
		header_color_str = fg(p_view_color_str)
		line_str         = line.strip().decode("utf-8")

		if "ERROR" in line_str or "error" in line_str:
			print("%s%s:%s%s%s%s"%(header_color_str, p_view_type_str, attr(0), bg("red"), line_str, attr(0)))
		else:
			print("%s%s:%s%s"%(header_color_str, p_view_type_str, attr(0), line_str))

	p_std_stream.close()

#---------------------------------------------------
# DEPRECATED!! - move all users over to run()

def run_cmd(p_cmd_str,
	p_env_map           = None,
	p_print_output_bool = True):
	
	if p_print_output_bool:
		print(p_cmd_str)
	
	if not p_env_map == None:
		assert isinstance(p_env_map, dict)
		r = delegator.run(p_cmd_str, env=p_env_map)
	else:
		r = delegator.run(p_cmd_str)

	o = ""
	e = ""

	# sometimes commands dont return any stdout
	if not r.out == "":
		o = r.out
		if p_print_output_bool:
			print(o)

	# sometimes commands dont return any stderr
	if not r.err == "":
		e = r.err
		if p_print_output_bool: print(e)
	
	return o, e, r.return_code

#-----------------------------------------------------
# DEPRECATED!! - all users of this function will be migrated to using Pythons stdlib argparse
#                directly. once they all migrate to this remove this function and this abstraction.
# IMPORTANT!! - these arguments(service_info parameters) have precendence over
#               what is returned by services get_service_info() function

def parse_args(p_cmd_line_args_defs_map, p_log_fun):
	assert isinstance(p_cmd_line_args_defs_map, dict)
	
	passed_in_args_lst = sys.argv[1:]
	p_log_fun("INFO", "passed in args:%s"%(passed_in_args_lst))
	
	# RawTextHelpFormatter - so that newlines in the help text are rendered when "-h" option 
	#                        is passed on the command line
	arg_parser = argparse.ArgumentParser(formatter_class = argparse.RawTextHelpFormatter)

	#load up all command line argument definitions
	for arg_name_str, arg_def_map in p_cmd_line_args_defs_map.items():
		arg_default  = arg_def_map["default"]
		arg_help_str = arg_def_map["help"]
		

		arg_parser.add_argument("-%s"%(arg_name_str), 
			action  = "store",
			default = arg_default,
			help    = arg_help_str)
	
	#:Namespace
	args_namespace = arg_parser.parse_args(passed_in_args_lst)
	
	#extracts command line arguments from args_namespace 
	#(only the expected arguments), and repacks them into dynamic_service_info_dict
	extracted_args_map = {}
	for arg_name_str,_ in p_cmd_line_args_defs_map.items():
		extracted_args_map[arg_name_str] = getattr(args_namespace,arg_name_str)
		
	return extracted_args_map

#-----------------------------------------------------
def confirm(p_prompt_str, p_resp=False):
	prompt_str = None
	if p_prompt_str is None:
		prompt_str = "Confirm"

	if p_resp:
		prompt_str = "%s %s|%s: "%(p_prompt_str, "y", "n")
	else:
		prompt_str = "%s %s|%s: "%(p_prompt_str, "n", "y")
		
	while True:

		answer_str = input(prompt_str)
		if not answer_str:
			return p_resp
		if answer_str not in ["y", "Y", "n", "N"]:
			print("please enter y or n.")
			continue
		if answer_str == "y" or answer_str == "Y":
			return True
		if answer_str == "n" or answer_str == "N":
			return False


# GloFlow application and media management/publishing platform
# Copyright (C) 2021 Ivan Trajkovic
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

import os, sys
modd_str = os.path.abspath(os.path.dirname(__file__)) # module dir

import argparse
import json

import svgwrite
from svgwrite import cm, mm
from svgwrite.container import SVG

#--------------------------------------------------
def main():
	
	print("PY_PLUGIN - PLOT_TX_TRACE")
	#----------------------------
	# INPUT
	args_map = parse_args()
	


	l = sys.stdin.readline()
	print(">>>>>>>>>>>>>>>>>>>>>>>>>>")
	# print(l.strip())



	tx_id_str    = args_map["tx_id_str"]
	tx_trace_map = json.loads(l)
	assert isinstance(tx_trace_map, dict)

	#----------------------------



	plot_y = 50

	dwg      = svgwrite.Drawing(filename=f"{modd_str}/test.svg", debug=True)
	plot_svg = SVG((50, plot_y))
	dwg.add(plot_svg)


	plot_width_mm_int = 500
	
	x_ops_base      = 10
	x_stack_base    = 30
	x_memory_base   = x_stack_base + 18 
	x_gas_cost_base = x_memory_base + 10

	#----------------------------
	# LEGEND
	legend_svg = SVG((50, 10))
	dwg.add(legend_svg)

	legend_svg.add(dwg.text(f"tx ID - {tx_id_str}", (0*mm, 4*mm), font_size=8))

	legend_svg.add(dwg.text(f"stack",  (x_stack_base*mm, plot_y-12), font_size=8))
	legend_svg.add(dwg.text(f"memory", ((x_memory_base-3)*mm, plot_y-12), font_size=8))
	legend_svg.add(dwg.text(f"gas cost", (x_gas_cost_base*mm, plot_y-12), font_size=8))

	#----------------------------



	i=0

	memory_ops_lst = []
	call_ops_lst   = []
	logs_ops_lst   = []

	for op_map in tx_trace_map["opcodes_lst"]:
		

		# print("--------------------")
		# print(op_map)

		op_str       = op_map["op_str"].strip()
		gas_cost_int = int(op_map["gas_cost_uint"])
		# print(f"{op_str}-{gas_cost_int}")



		stack_lst  = op_map["stack_lst"]
		memory_lst = op_map["memory_lst"]



		

		x1 = x_gas_cost_base
		x2 = x1+gas_cost_int
		y  = i*8 # 2.4
		

		op_svg = SVG((0, y))
		plot_svg.add(op_svg)

		#----------------------------
		# OP_GAS_COST__LINE
		op_svg.add(dwg.line(start=(x1*mm, 1.25*mm), end=(x2*mm, 1.25*mm),
			stroke="green",
			stroke_width=3))

		#----------------------------
		# DEBUGGING - alignment line. used to align other elements in line.
		op_svg.add(dwg.line(start=((x_stack_base-1)*mm, 1.25*mm), end=(x1*mm, 1.25*mm),
			stroke="black",
			stroke_width=0.5))

		#----------------------------
		# OP__TEXT - text local coordinate system is at lower left corner,
		#            not upper-left like everything else.
		#            so positioning it a bit lower from 0,0 (in this case 0,2)
		op_txt = dwg.text(op_str, (x_ops_base*mm, 2*mm),
			font_size=8)
		op_svg.add(op_txt)

		
		# MSTORE
		if op_str == "MSTORE":
			y__local = 0.8
			op_svg.add(dwg.rect(insert=((x_ops_base-1.5)*mm, y__local*mm), size=(1*mm, 1*mm),
				fill='blue',
				stroke='black',
				stroke_width=0.5))


			y__global = y+y__local
			memory_ops_lst.append(y__global)


		# CALLDATASIZE/CALLVALUE/CALLER
		ops_call_lst = ["CALLDATASIZE", "CALLVALUE", "CALLER"]
		if op_str in ops_call_lst:
			
			x_call_rect__global = x_ops_base-2
			y__local            = 0.6
			op_svg.add(dwg.rect(insert=((x_call_rect__global)*mm, y__local*mm), size=(1.2*mm, 1.2*mm),
				fill='red',
				stroke='black',
				stroke_width=0.5))

			y__global = y+y__local
			call_ops_lst.append(y__global)

		# LOG
		if op_str.startswith("LOG"):
			
			x_log_rect__global = x_ops_base-2
			y__local = 0.6
			op_svg.add(dwg.rect(insert=(x_log_rect__global*mm, y__local*mm), size=(1*mm, 1*mm),
				fill='cyan',
				stroke='black',
				stroke_width=0.5))
			
			y__global = y+y__local
			logs_ops_lst.append(y__global)

		#----------------------------
		# STACK
		op_stack_g = op_svg.add(dwg.g(id='op_stack', stroke='blue'))
		j=0
		for s in stack_lst:

			x = x_stack_base + j*2
			op_stack_g.add(dwg.rect(insert=(x*mm, 0.5*mm), size=(1.5*mm, 1.5*mm),
				fill='yellow',
				stroke='black',
				stroke_width=0.5))

			j+=1

		#----------------------------
		# MEMORY
		op_memory_g = op_svg.add(dwg.g(id='op_memory', stroke='blue'))
		j=0
		for s in memory_lst:

			x = x_memory_base + j*2
			op_memory_g.add(dwg.rect(insert=(x*mm, 0.5*mm), size=(1.5*mm, 1.5*mm),
				fill='orange',
				stroke='black',
				stroke_width=0.5))

			j+=1

		#----------------------------

		i+=1

	
	#--------------------------------------------------
	# ARCHS
	def draw_archs():
		
		x_call_rect__global_px = 30

		i=0
		for y__global in call_ops_lst[:-1]:

			y__global__next = call_ops_lst[i+1]

			y__global__start_str = f"{x_call_rect__global_px} {int(y__global+4.4)}"
			y__global__end_str   = f"{x_call_rect__global_px} {int(y__global__next+4.4)}"
			y__global__control_point_str = f"0 {int(y__global+(y__global__next-y__global)/2)}"

			path_str = f"M {y__global__start_str} Q {y__global__control_point_str} {y__global__end_str}"

			plot_svg.add(dwg.path(d=path_str, fill="none", stroke="red", stroke_width=0.5))
			
			i+=1

	#--------------------------------------------------
	draw_archs()

	print("done drawing...")

	# FILE_SAVE
	if args_map["stdout_bool"]:

		svg_str = dwg.tostring()
		out_map = {"svg_str": svg_str}
		print(f"GF_OUT:{json.dumps(out_map)}")
	else:
		dwg.save()

#--------------------------------------------------
def parse_args():
	arg_parser = argparse.ArgumentParser(formatter_class = argparse.RawTextHelpFormatter)
	#----------------------------
	# TX_ID
	arg_parser.add_argument("-tx_id", action = "store", default=None,
		help = "hex ID of the target transaction")

	#----------------------------
	# STDOUT
	arg_parser.add_argument('-stdout', action='store_true')

	#----------------------------

	cli_args_lst   = sys.argv[1:]
	args_namespace = arg_parser.parse_args(cli_args_lst)

	return {
		"tx_id_str":   args_namespace.tx_id,
		"stdout_bool": args_namespace.stdout
	}

#--------------------------------------------------
main()


# GloFlow application and media management/publishing platform
# Copyright (C) 2019 Ivan Trajkovic
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

import os, sys
modd_str = os.path.abspath(os.path.dirname(__file__)) # module dir

from colored import fg, bg, attr
import delegator

sys.path.append("%s/../../gf_core"%(modd_str))
import gf_core_cli

#--------------------------------------------------
# p_mark_all_bool - mark all apps as changed. used mainly for debugging.

def list_changed_apps(p_apps_changes_deps_map,
    p_commits_lookback_int = 1,
    p_mark_all_bool        = False):
    assert isinstance(p_apps_changes_deps_map, dict)
    assert "apps_gf_packages_map" in p_apps_changes_deps_map.keys()
    assert "system_packages_lst" in p_apps_changes_deps_map.keys()
    assert isinstance(p_commits_lookback_int, int)

    apps_gf_packages_map = p_apps_changes_deps_map['apps_gf_packages_map']
    system_packages_lst  = p_apps_changes_deps_map['system_packages_lst']
    assert isinstance(system_packages_lst, list)

    changed_apps_files_map  = {
        # IMPORTANT!! - these are all apps that have either "go" or "web" changed. this is 
        #               needed because when building in CI
        #               even if only Go files changed we need Web code built as well so that
        #               the final container can be built in its full form.
        "all": {},
        "go":  {},
        "web": {},
    }

    #------------------------
    # DEBUGGING - mark all apps as changed
    if p_mark_all_bool:
        for a, _ in apps_gf_packages_map.items():
            changed_apps_files_map["all"][a] = ["all"]
            changed_apps_files_map["go"][a]  = ["all"]
            changed_apps_files_map["web"][a] = ["all"]
        return changed_apps_files_map

    #------------------------

    # latest_commit_hash_str, _ = gf_core_cli.run('git rev-parse HEAD')
    # assert len(latest_commit_hash_str) == 32

    #------------------------
    # FIX!! - dont just look 1 commit back to see what changed. if localy a developer makes several commits and then uploads code
    #         got github(or other) and CI clones it this function then might miss some of the services/apps/packages that changed 
    #         several commits back.
    #         instead some mechanism for getting the number of commits that some deployment environment is behind HEAD,
    #         and then use that number for this "p_commits_lookback_int" argument (that would be >1).
    past_commit_str = "HEAD~%s"%(p_commits_lookback_int)

    #------------------------

    list_str, _, _ = gf_core_cli.run("git diff --name-only HEAD %s"%(past_commit_str), p_print_output_bool=False)

    #--------------------------------------------------
    # IMPORTANT!! - the file that changed affects all apps, so they all need to be marked as changed
    #               and this file added to the list of changed files of all apps.

    def add_change_to_all_apps(p_file_changed_str, p_type_str):
        assert p_type_str == "go" or p_type_str == "web"
        for a, _ in apps_gf_packages_map.items():

            if a in changed_apps_files_map[p_type_str].keys():
                changed_apps_files_map["all"][a].append(p_file_changed_str)
                changed_apps_files_map[p_type_str][a].append(p_file_changed_str)
            else:
                changed_apps_files_map["all"][a].append(p_file_changed_str)
                changed_apps_files_map[p_type_str][a] = [p_file_changed_str]

    #--------------------------------------------------
    # IMPORTANT!! - update only the apps that have this files package is marked as a dependancy of
    def update_dependant_apps_file_lists(p_package_name_str, p_file_path_str, p_type_str):
        assert p_type_str == "go" or p_type_str == "web"

        # build out a list of apps that this package (p_package_name_str) is a dependency of
        dependant_apps_lst = []
        for app_str, app_gf_package_lst in p_apps_changes_deps_map['apps_gf_packages_map'].items():
            if p_package_name_str in app_gf_package_lst:
                dependant_apps_lst.append(app_str)

        # for all apps that are determined to have changed (because they depend on p_package_name_str package) 
        # add this file (p_file_path_str) to those apps lists of changed files.
        for app_str in dependant_apps_lst:
            
            if app_str in changed_apps_files_map.keys():
                changed_apps_files_map["all"][app_str].append(p_file_changed_str)
                changed_apps_files_map[p_type_str][app_str].append(p_file_changed_str)
            else:  
                changed_apps_files_map["all"][app_str].append(p_file_changed_str)                                     
                changed_apps_files_map[p_type_str][app_str] = [p_file_changed_str]

    #--------------------------------------------------
    
    for l in list_str.split('\n'):

        #------------------------
        # GO
        if l.startswith('go'):
            # an app changed
            if l.startswith('go/gf_apps'):
                package_name_str = l.split('/')[2] # third element in the file path is a package name
                update_dependant_apps_file_lists(package_name_str, l, "go")

            # one of the system packages has changed
            else:
                for sys_package_str in system_packages_lst:

                    # IMPORTANT!! - one of the system packages has changed, so infer
                    #              that all apps have changed.
                    if l.startswith('go/%s'%(sys_package_str)):
                        add_change_to_all_apps(l, "go")

        #------------------------
        # WEB
        elif l.startswith('web'):
            if l.startswith('web/src/gf_apps'):
                package_name_str = l.split('/')[3] # get package_name from the path of the changed file
                update_dependant_apps_file_lists(package_name_str, l, "web")

            # IMPORTATN!! - one of the web libs changed, so all apps should be rebuilt
            # FIX!!       - have a better way of determening which apps use this lib, 
            #               to avoid rebuilding unaffected apps
            elif l.startswith('web/libs'):
                add_change_to_all_apps(l, "web")

            else:
                package_name_str = l.split('/')[2]
                if package_name_str in system_packages_lst:
                    add_change_to_all_apps(l, web)

        #------------------------

    return changed_apps_files_map

#--------------------------------------------------
def view_changed_apps(p_changed_apps_files_map, p_type_str):
    assert isinstance(p_changed_apps_files_map, dict)
    assert p_type_str == "go" or p_type_str == "web"

    print("----- %s"%(p_type_str))

    if len(p_changed_apps_files_map[p_type_str].items()) == 0:
        print('NO APPS CHANGED')
    else:
        for app_name_str, changed_files_lst in p_changed_apps_files_map[p_type_str].items():
            print('%s%s%s'%(fg('yellow'), app_name_str, attr(0)))
            
            if len(changed_files_lst) == 0:
                print('%s%s%s'%(fg('green'),
                    "ALL",
                    attr(0)))

            for f in changed_files_lst:

                print('\t%s/%s%s%s'%(os.path.dirname(f),
                    fg('green'),
                    os.path.basename(f),
                    attr(0)))